# coding: utf-8

"""
    Tator REST API

    Interface to the Tator backend.  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from ..configuration import Configuration


class MediaFiles(object):
    """
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'archival': 'list[VideoDefinition]',
        'attachment': 'list[AuxiliaryFileDefinition]',
        'audio': 'list[AudioDefinition]',
        'concat': 'list[ConcatDefinition]',
        'frame_offset': 'list[int]',
        'ids': 'list[int]',
        'image': 'list[ImageDefinition]',
        'layout': 'list[int]',
        'live': 'list[LiveDefinition]',
        'quality': 'int',
        'streaming': 'list[VideoDefinition]',
        'thumbnail': 'list[ImageDefinition]',
        'thumbnail_gif': 'list[ImageDefinition]'
    }

    attribute_map = {
        'archival': 'archival',
        'attachment': 'attachment',
        'audio': 'audio',
        'concat': 'concat',
        'frame_offset': 'frameOffset',
        'ids': 'ids',
        'image': 'image',
        'layout': 'layout',
        'live': 'live',
        'quality': 'quality',
        'streaming': 'streaming',
        'thumbnail': 'thumbnail',
        'thumbnail_gif': 'thumbnail_gif'
    }

    def __init__(self, archival=None, attachment=None, audio=None, concat=None, frame_offset=None, ids=None, image=None, layout=None, live=None, quality=None, streaming=None, thumbnail=None, thumbnail_gif=None, local_vars_configuration=None):  # noqa: E501
        """MediaFiles - a model defined in OpenAPI"""
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._archival = None
        self._attachment = None
        self._audio = None
        self._concat = None
        self._frame_offset = None
        self._ids = None
        self._image = None
        self._layout = None
        self._live = None
        self._quality = None
        self._streaming = None
        self._thumbnail = None
        self._thumbnail_gif = None
        self.discriminator = None

        if archival is not None:
            self.archival = archival
        if attachment is not None:
            self.attachment = attachment
        if audio is not None:
            self.audio = audio
        if concat is not None:
            self.concat = concat
        if frame_offset is not None:
            self.frame_offset = frame_offset
        if ids is not None:
            self.ids = ids
        if image is not None:
            self.image = image
        if layout is not None:
            self.layout = layout
        if live is not None:
            self.live = live
        if quality is not None:
            self.quality = quality
        if streaming is not None:
            self.streaming = streaming
        if thumbnail is not None:
            self.thumbnail = thumbnail
        if thumbnail_gif is not None:
            self.thumbnail_gif = thumbnail_gif

    @property
    def archival(self):
        """

        :return: The archival of this MediaFiles. 
        :rtype: list[VideoDefinition]
        """
        return self._archival

    @archival.setter
    def archival(self, archival):
        """

        :param archival: The archival of this MediaFiles.
        :type: list[VideoDefinition]
        """

        self._archival = archival

    @property
    def attachment(self):
        """

        :return: The attachment of this MediaFiles. 
        :rtype: list[AuxiliaryFileDefinition]
        """
        return self._attachment

    @attachment.setter
    def attachment(self, attachment):
        """

        :param attachment: The attachment of this MediaFiles.
        :type: list[AuxiliaryFileDefinition]
        """

        self._attachment = attachment

    @property
    def audio(self):
        """

        :return: The audio of this MediaFiles. 
        :rtype: list[AudioDefinition]
        """
        return self._audio

    @audio.setter
    def audio(self, audio):
        """

        :param audio: The audio of this MediaFiles.
        :type: list[AudioDefinition]
        """

        self._audio = audio

    @property
    def concat(self):
        """

        :return: The concat of this MediaFiles. 
        :rtype: list[ConcatDefinition]
        """
        return self._concat

    @concat.setter
    def concat(self, concat):
        """

        :param concat: The concat of this MediaFiles.
        :type: list[ConcatDefinition]
        """

        self._concat = concat

    @property
    def frame_offset(self):
        """
        Frame of sub-video, offset from media in slot 0.

        :return: The frame_offset of this MediaFiles. 
        :rtype: list[int]
        """
        return self._frame_offset

    @frame_offset.setter
    def frame_offset(self, frame_offset):
        """
        Frame of sub-video, offset from media in slot 0.

        :param frame_offset: The frame_offset of this MediaFiles.
        :type: list[int]
        """

        self._frame_offset = frame_offset

    @property
    def ids(self):
        """
        If multi-stream list of ids of sub-videos

        :return: The ids of this MediaFiles. 
        :rtype: list[int]
        """
        return self._ids

    @ids.setter
    def ids(self, ids):
        """
        If multi-stream list of ids of sub-videos

        :param ids: The ids of this MediaFiles.
        :type: list[int]
        """

        self._ids = ids

    @property
    def image(self):
        """

        :return: The image of this MediaFiles. 
        :rtype: list[ImageDefinition]
        """
        return self._image

    @image.setter
    def image(self, image):
        """

        :param image: The image of this MediaFiles.
        :type: list[ImageDefinition]
        """

        self._image = image

    @property
    def layout(self):
        """
        2-element array to define rxc layout

        :return: The layout of this MediaFiles. 
        :rtype: list[int]
        """
        return self._layout

    @layout.setter
    def layout(self, layout):
        """
        2-element array to define rxc layout

        :param layout: The layout of this MediaFiles.
        :type: list[int]
        """

        self._layout = layout

    @property
    def live(self):
        """

        :return: The live of this MediaFiles. 
        :rtype: list[LiveDefinition]
        """
        return self._live

    @live.setter
    def live(self, live):
        """

        :param live: The live of this MediaFiles.
        :type: list[LiveDefinition]
        """

        self._live = live

    @property
    def quality(self):
        """
        Resolution to fetch on each sub-video

        :return: The quality of this MediaFiles. 
        :rtype: int
        """
        return self._quality

    @quality.setter
    def quality(self, quality):
        """
        Resolution to fetch on each sub-video

        :param quality: The quality of this MediaFiles.
        :type: int
        """

        self._quality = quality

    @property
    def streaming(self):
        """

        :return: The streaming of this MediaFiles. 
        :rtype: list[VideoDefinition]
        """
        return self._streaming

    @streaming.setter
    def streaming(self, streaming):
        """

        :param streaming: The streaming of this MediaFiles.
        :type: list[VideoDefinition]
        """

        self._streaming = streaming

    @property
    def thumbnail(self):
        """

        :return: The thumbnail of this MediaFiles. 
        :rtype: list[ImageDefinition]
        """
        return self._thumbnail

    @thumbnail.setter
    def thumbnail(self, thumbnail):
        """

        :param thumbnail: The thumbnail of this MediaFiles.
        :type: list[ImageDefinition]
        """

        self._thumbnail = thumbnail

    @property
    def thumbnail_gif(self):
        """

        :return: The thumbnail_gif of this MediaFiles. 
        :rtype: list[ImageDefinition]
        """
        return self._thumbnail_gif

    @thumbnail_gif.setter
    def thumbnail_gif(self, thumbnail_gif):
        """

        :param thumbnail_gif: The thumbnail_gif of this MediaFiles.
        :type: list[ImageDefinition]
        """

        self._thumbnail_gif = thumbnail_gif

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, MediaFiles):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, MediaFiles):
            return True

        return self.to_dict() != other.to_dict()



""" Module for handling the boiler plate around adding media resources to existing Media objects """

from .md5sum import md5sum
from ..transcode.transcode import make_video_definition, convert_streaming, convert_archival
import tempfile
import os


def supplement_video_to_media(api,
                              media_id,
                              role,
                              configuration,
                              filepath: str,
                              force=False):
  """ This routine will transcode `filepath` based on `configuration` to `media_object`

  See :py:meth:`tator.api.delete_video_file` for information on deleting resolutions from a media object  

  :param api: An initialized tator.api project
  :param media_id: Id of the media to supplement to
  :param role: One of 'streaming' or 'archival'
  :param configuration: For 'streaming' roles, must be a tator.models.ResolutionConfig, for 'archival' roles must be a tator.models.EncodeConfig
  :param filepath: The local file path to the file to encode
  :param force: If true, will ignore any differences in the source file compared to what was previously encoded. This is very dangerous as it could result in streams of different lengths or even different files.
  
  """
  media_object = api.get_media(media_id)
  media_type_object = api.get_media_type(media_object.type)
  if media_type_object.dtype != 'video':
    raise Exception(f'{media_id} is not a video, its a {media_type_object.dtype}')
  
  if not role in ['streaming', 'archival']:
    raise Exception(f'"{role}" must be one of "streaming" or "archival"')

  md5_of_file = md5sum(filepath)

  # Do not proceed if force is false + the input file doesn't appear to match the original inputted file
  if md5_of_file != media_object.md5 and force is False:
    raise Exception(f'{media_id} has an md5 of {media_object.md5}, local file is {md5_of_file}. Need to supply force=True to bypass this safety check.')
  else:
    print("WARNING: Bypassing source uniqueness check")

  # Extract host + token from supplied api object.
  host = api.api_client.configuration.host
  token = api.api_client.configuration.api_key['Authorization']

  # Construct the raw definition to grab input file properties
  raw_definition = make_video_definition(filepath)
  height, width = raw_definition['resolution']

  if role == 'streaming':
    # In a temporary directory do all the work
    with tempfile.TemporaryDirectory() as td:
      # To be compatible with the convert_streaming routine need to stringify the configuration
      resolution = configuration.resolution
      crf = configuration.crf if configuration.crf else 23
      codec = configuration.vcodec if configuration.vcodec else 'libx264'
      preset = configuration.preset if configuration.preset else ''
      configs = [f"{resolution}:{crf}:{codec}:{preset}"]
      convert_streaming(host, token, media_id, filepath, td, width, height, configs)

      

  if role == 'archival':
    with tempfile.TemporaryDirectory() as td:
      archival_path = os.path.join(td, "archival.mp4")
      convert_archival(host, token, media_id, filepath, archival_path, width, height, None, [configuration])



import math
import os
import requests
import subprocess
import tempfile
import logging
from uuid import uuid1

from PIL import Image

from .md5sum import md5sum
from .download_media import download_media
from ._upload_file import _upload_file
from ._download_file import _download_file
from ..openapi.tator_openapi.models import MessageResponse

logger = logging.getLogger(__name__)

def make_concat(api, name, media_ids, section, offsets=None):
    """ Uploads a single media file.

    :param api: :class:`tator.TatorApi` object.
    :param name: Name of the file to use
    :param media_ids: List of media_ids to multi-stream
    :param section: Section name. If this section does not exist it will be created.
    :param offsets: Section name. If this section does not exist it will be created.
    :returns: Response from media object creation.
    """
    media_obj = api.get_media(media_ids[0], presigned=86400)
    type_id = media_obj.type;
    project = media_obj.project;
    # if offsets aren't supplied make them
    all_media = api.get_media_list_by_id(project, {"ids": media_ids})

    # use a concat extension
    name += ".concat"
   
    # Fetch the section. If it does not exist, create it. 
    sections = api.get_section_list(project, name=section)
    if len(sections) == 0:
        section_spec = {'name': section, 'tator_user_sections': str(uuid1())}
        response = api.create_section(project, section_spec=section_spec)
        section_obj = api.get_section_list(project, name=section)[0]
    else:
        section_obj = sections[0]


    attributes={}
    if section:
        attributes.update({"tator_user_sections": section_obj.tator_user_sections})

    if offsets == None:
        offsets = [0]
        for media in all_media[0:len(all_media)-1]:
            offsets.append(media.num_frames / media.fps)


    md5 = media_obj.md5 # doesn't matter
    media_spec = {'attributes': attributes,
                  'name': name,
                  'md5': md5,
                  'section': section_obj.name,
                  'fps': media_obj.fps,
                  'num_frames': media_obj.num_frames,
                  'type': type_id,
                  'width': media_obj.width,
                  'height': media_obj.height}

    resp = api.create_media_list(project, [media_spec])
    print(f"Created {resp.id[0]}")

    # Copy thumbnails from first media
    with tempfile.TemporaryDirectory() as d:
        thumb_path = os.path.join(d,'tiled_thumb.jpg')
        thumb_gif_path = os.path.join(d,'tiled_gif.gif')

        for progress in download_media(api, media_obj, thumb_path, None, 'thumbnail'):
            pass
        for progress in download_media(api, media_obj, thumb_gif_path, None, 'thumbnail_gif'):
            pass

        for progress, thumbnail_info in _upload_file(api, project, thumb_path,
                                                     media_id=resp.id[0], filename='tiled_thumb.jpg'):
            logger.info(f"Thumbnail upload progress: {progress}%")
        for progress, thumbnail_gif_info in _upload_file(api, project, thumb_gif_path,
                                                         media_id=resp.id[0], filename='tiled_gif.gif'):
            logger.info(f"Thumbnail gif upload progress: {progress}%")

        # Open images to get output resolution.
        thumb_image = Image.open(thumb_path)
        thumb_gif_image = Image.open(thumb_gif_path)

        # Create image definitions for thumbnails.
        thumb_def = {'path': thumbnail_info.key,
                     'size': os.stat(thumb_path).st_size,
                     'resolution': [thumb_image.height, thumb_image.width],
                     'mime': f'image/{thumb_image.format.lower()}'}
        thumb_gif_def = {'path': thumbnail_gif_info.key,
                         'size': os.stat(thumb_gif_path).st_size,
                         'resolution': [thumb_gif_image.height, thumb_gif_image.width],
                         'mime': f'image/{thumb_gif_image.format.lower()}'}

        response = api.create_image_file(resp.id[0], role='thumbnail', image_definition=thumb_def)
        assert isinstance(response, MessageResponse)
        response = api.create_image_file(resp.id[0], role='thumbnail_gif', image_definition=thumb_gif_def)
        assert isinstance(response, MessageResponse)

    concat_def = [{"id": media_id, "timestampOffset": offset} for media_id,offset in zip(media_ids, offsets)]
    api.update_media(resp.id[0], {"concat": concat_def})

    return resp




# coding: utf-8

"""
    Tator REST API

    Interface to the Tator backend.  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from ..configuration import Configuration


class Organization(object):
    """
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'default_membership_permission': 'str',
        'effective_permission': 'int',
        'id': 'int',
        'name': 'str',
        'permission': 'str',
        'thumb': 'str'
    }

    attribute_map = {
        'default_membership_permission': 'default_membership_permission',
        'effective_permission': 'effective_permission',
        'id': 'id',
        'name': 'name',
        'permission': 'permission',
        'thumb': 'thumb'
    }

    def __init__(self, default_membership_permission=None, effective_permission=None, id=None, name=None, permission=None, thumb=None, local_vars_configuration=None):  # noqa: E501
        """Organization - a model defined in OpenAPI"""
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._default_membership_permission = None
        self._effective_permission = None
        self._id = None
        self._name = None
        self._permission = None
        self._thumb = None
        self.discriminator = None

        if default_membership_permission is not None:
            self.default_membership_permission = default_membership_permission
        if effective_permission is not None:
            self.effective_permission = effective_permission
        if id is not None:
            self.id = id
        if name is not None:
            self.name = name
        if permission is not None:
            self.permission = permission
        if thumb is not None:
            self.thumb = thumb

    @property
    def default_membership_permission(self):
        """
        Default user permission level for all projects in this organization. If specified, users in this organizaiton will be automatically added to all projects in this organization with at least this permission level.

        :return: The default_membership_permission of this Organization. 
        :rtype: str
        """
        return self._default_membership_permission

    @default_membership_permission.setter
    def default_membership_permission(self, default_membership_permission):
        """
        Default user permission level for all projects in this organization. If specified, users in this organizaiton will be automatically added to all projects in this organization with at least this permission level.

        :param default_membership_permission: The default_membership_permission of this Organization.
        :type: str
        """
        allowed_values = ["No Access", "View Only", "Can Edit", "Can Transfer", "Can Execute", "Full Control"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and default_membership_permission not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `default_membership_permission` ({0}), must be one of {1}"  # noqa: E501
                .format(default_membership_permission, allowed_values)
            )

        self._default_membership_permission = default_membership_permission

    @property
    def effective_permission(self):
        """
        The effective permission for the user making the request

        :return: The effective_permission of this Organization. 
        :rtype: int
        """
        return self._effective_permission

    @effective_permission.setter
    def effective_permission(self, effective_permission):
        """
        The effective permission for the user making the request

        :param effective_permission: The effective_permission of this Organization.
        :type: int
        """

        self._effective_permission = effective_permission

    @property
    def id(self):
        """
        Unique integer identifying the organization.

        :return: The id of this Organization. 
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """
        Unique integer identifying the organization.

        :param id: The id of this Organization.
        :type: int
        """

        self._id = id

    @property
    def name(self):
        """
        Name of the organization.

        :return: The name of this Organization. 
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Name of the organization.

        :param name: The name of this Organization.
        :type: str
        """

        self._name = name

    @property
    def permission(self):
        """
        Permission level of user making request.

        :return: The permission of this Organization. 
        :rtype: str
        """
        return self._permission

    @permission.setter
    def permission(self, permission):
        """
        Permission level of user making request.

        :param permission: The permission of this Organization.
        :type: str
        """

        self._permission = permission

    @property
    def thumb(self):
        """
        S3 key of thumbnail used to represent the organization.

        :return: The thumb of this Organization. 
        :rtype: str
        """
        return self._thumb

    @thumb.setter
    def thumb(self, thumb):
        """
        S3 key of thumbnail used to represent the organization.

        :param thumb: The thumb of this Organization.
        :type: str
        """

        self._thumb = thumb

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Organization):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, Organization):
            return True

        return self.to_dict() != other.to_dict()



from datetime import datetime
from functools import partial
from pprint import pprint
import pytest
import random
import string
import uuid

import tator


def random_state(project, state_type, video_obj, post=False):
    attributes = {
        "test_bool": random.choice([False, True]),
        "test_int": random.randint(-1000, 1000),
        "test_float": random.uniform(-1000.0, 1000.0),
        "test_enum": random.choice(["a", "b", "c"]),
        "test_string": str(uuid.uuid1()),
        "test_datetime": datetime.now().isoformat(),
        "test_geopos": [random.uniform(-180.0, 180.0), random.uniform(-90.0, 90.0)],
        "test_float_array": [random.uniform(-1.0, 1.0) for _ in range(3)],
    }
    out = {
        "project": project,
        "type": state_type,
        "media_ids": [video_obj.id],
        "frame": random.randint(0, video_obj.num_frames - 1),
        "attributes": attributes,
    }

    return {**out}


def random_localization(project, box_type, video_obj, post=False):
    x = random.uniform(0.0, 1.0)
    y = random.uniform(0.0, 1.0)
    w = random.uniform(0.0, 1.0 - x)
    h = random.uniform(0.0, 1.0 - y)
    attributes = {
        "test_bool": random.choice([False, True]),
        "test_int": random.randint(-1000, 1000),
        "test_float": random.uniform(-1000.0, 1000.0),
        "test_enum": random.choice(["a", "b", "c"]),
        "test_string": str(uuid.uuid4()),
        "test_datetime": datetime.now().isoformat(),
        "test_geopos": [random.uniform(-180.0, 180.0), random.uniform(-90.0, 90.0)],
        "test_float_array": [random.uniform(-1.0, 1.0) for _ in range(3)],
    }
    out = {
        "x": x,
        "y": y,
        "width": w,
        "height": h,
        "project": project,
        "type": box_type,
        "media_id": video_obj.id,
        "frame": random.randint(0, video_obj.num_frames - 1),
        "attributes": attributes,
    }

    return {**out}


def random_leaf(project, leaf_type, parent_obj=None, post=False):
    attributes = {
        "test_bool": random.choice([False, True]),
        "test_int": random.randint(-1000, 1000),
        "test_float": random.uniform(-1000.0, 1000.0),
        "test_enum": random.choice(["a", "b", "c"]),
        "test_string": str(uuid.uuid1()),
        "test_datetime": datetime.now().isoformat(),
        "test_geopos": [random.uniform(-180.0, 180.0), random.uniform(-90.0, 90.0)],
        "test_float_array": [random.uniform(-1.0, 1.0) for _ in range(3)],
    }
    name = "".join(random.choice(string.ascii_letters) for _ in range(10))
    out = {"project": project, "type": leaf_type, "name": name, "attributes": attributes}
    if parent_obj:
        out["parent"] = parent_obj.id

    return {**out}


def compare_change_logs(new, old):
    for old_in_new in new.description_of_change.old:
        for new_in_old in old.description_of_change.new:
            if new_in_old.name == old_in_new.name:
                if type(new_in_old.value) == float or type(old_in_new.value) == float:
                    is_equal = new_in_old.value == pytest.approx(old_in_new.value)
                else:
                    is_equal = new_in_old.value == old_in_new.value

                assert (
                    is_equal
                ), f"FAILED '{new_in_old.name}': {new_in_old.value} != {old_in_new.value}"


def change_log_helper(
    tator_api,
    random_entity,
    entity_type,
    project,
    create_list,
    update_one,
    update_list,
    delete_one,
    delete_list,
):
    num_entities = 2
    entities = [random_entity(post=True) for _ in range(num_entities)]
    response = create_list(project, entities)
    entity_ids = response.id

    assert len(entity_ids) == len(entities)

    # Creation tests
    create_changes = []
    for entity_id, box in zip(entity_ids, entities):
        changes = tator_api.get_change_log_list(
            project=project, entity_id=entity_id, entity_type=entity_type
        )

        # Assert one change returned
        assert len(changes) == 1
        changes = changes[0]

        # Assert all old values are None for creation
        for change in changes.description_of_change.old:
            assert change.value == None

        # Assert all new values match initial creation values
        for change in changes.description_of_change.new:
            if change.name == "_id":
                assert change.value == entity_id
            elif change.name in ["_x", "_y", "_width", "_height", "_frame"]:
                assert change.value == box[change.name.replace("_", "")]
            elif change.name.startswith("test_"):
                assert change.value == box["attributes"][change.name]
        create_changes.append(changes)

    patch_entities = [random_entity() for _ in range(num_entities)]
    for p in patch_entities:
        p["in_place"] = 1
    for entity_id, patch_entity in zip(entity_ids, patch_entities):
        update_one(entity_id, patch_entity)

    # Update tests
    patch_changes = []
    for entity_id, box, old_change_log in zip(entity_ids, patch_entities, create_changes):
        changes = tator_api.get_change_log_list(
            project=project, entity_id=entity_id, entity_type=entity_type
        )

        # Assert two changes returned
        assert len(changes) == 2
        for new_change_log in changes:
            if new_change_log.id != old_change_log.id:
                break
        else:
            assert False, "No new change log detected"

        # Assert all old values are from creation
        compare_change_logs(new_change_log, old_change_log)

        # Assert all new values match initial creation values
        for change in new_change_log.description_of_change.new:
            if change.name == "_id":
                assert change.value == entity_id
            elif change.name in ["_x", "_y", "_width", "_height", "_frame"]:
                assert change.value == box[change.name.replace("_", "")]
            elif change.name.startswith("test_"):
                assert change.value == box["attributes"][change.name]
        patch_changes.append(new_change_log)

    bulk_patch_entity = random_entity()
    bulk_patch_entity = {"attributes": bulk_patch_entity["attributes"], "in_place": 1}
    update_list(bulk_patch_entity)

    # Bulk update tests
    bulk_patch_changes = []
    for entity_id, create_change_log, patch_change_log in zip(
        entity_ids, create_changes, patch_changes
    ):
        changes = tator_api.get_change_log_list(
            project=project, entity_id=entity_id, entity_type=entity_type
        )

        # Assert two changes returned
        assert len(changes) == 3
        for new_change_log in changes:
            if new_change_log.id not in [patch_change_log.id, create_change_log.id]:
                break
        else:
            assert False, "No new change log detected"

        # Assert all new values match bulk update values
        for change in new_change_log.description_of_change.new:
            if change.name == "_id":
                assert change.value == entity_id
            elif change.name in ["_x", "_y", "_width", "_height", "_frame"]:
                assert change.value == bulk_patch_entity[change.name.replace("_", "")]
            elif change.name.startswith("test_"):
                assert change.value == bulk_patch_entity["attributes"][change.name]
        bulk_patch_changes.append(new_change_log)

    # Clean up
    delete_one(entity_ids[0])
    delete_list()


def test_attribute_box_type_change_log(host, token, project, attribute_video, attribute_box_type):
    tator_api = tator.get_api(host, token)
    video_obj = tator_api.get_media(attribute_video)
    random_entity = partial(random_localization, project, attribute_box_type, video_obj)
    create_list = tator_api.create_localization_list
    update_one = tator_api.update_localization
    update_list = partial(tator_api.update_localization_list, project, media_id=[attribute_video])
    params = {"media_id": [attribute_video], "type": attribute_box_type}
    delete_one = tator_api.delete_localization
    delete_list = partial(tator_api.delete_localization_list, project, **params)
    change_log_helper(
        tator_api,
        random_entity,
        "localization",
        project,
        create_list,
        update_one,
        update_list,
        delete_one,
        delete_list,
    )


def test_state_type_change_log(host, token, project, attribute_video, state_type):
    tator_api = tator.get_api(host, token)
    video_obj = tator_api.get_media(attribute_video)
    random_entity = partial(random_state, project, state_type, video_obj)
    create_list = tator_api.create_state_list
    update_one = tator_api.update_state
    update_list = partial(tator_api.update_state_list, project, media_id=[attribute_video])
    params = {"media_id": [attribute_video], "type": state_type}
    delete_one = tator_api.delete_state
    delete_list = partial(tator_api.delete_state_list, project, **params)
    change_log_helper(
        tator_api,
        random_entity,
        "state",
        project,
        create_list,
        update_one,
        update_list,
        delete_one,
        delete_list,
    )


def test_media_change_log(host, token, project, attribute_video_type):
    tator_api = tator.get_api(host, token)

    # Define media spec.
    num_media = 3
    fname = "MediaChangeLogTest.mp4"
    media_specs = [
        {
            "type": attribute_video_type,
            "uid": str(uuid.uuid1()),
            "gid": str(uuid.uuid1()),
            "name": fname,
            "md5": str(uuid.uuid1())[:32],
            "section": "Test media change log",
            "attributes": {"test_int": random.randint(0, 100)},
        }
        for _ in range(num_media)
    ]

    # Create the media.
    media_ids = [
        tator_api.create_media_list(project=project, body=[media_spec]).id[0]
        for media_spec in media_specs
    ]

    # Creation tests
    create_changes = []
    for media_id, media_spec in zip(media_ids, media_specs):
        changes = tator_api.get_change_log_list(
            project=project, entity_id=media_id, entity_type="media"
        )

        # Assert one change returned
        assert len(changes) == 1
        changes = changes[0]

        # Assert all old values are None for creation
        for change in changes.description_of_change.old:
            assert change.value == None

        # Assert all new values match initial creation values
        for change in changes.description_of_change.new:
            if change.name == "_id":
                assert change.value == media_id
            elif change.name in ["_type", "_uid", "_gid", "_name", "_md5", "_section"]:
                assert change.value == media_spec[change.name.replace("_", "")]

        create_changes.append(changes)

    media_updates = [{"attributes": {"test_int": random.randint(0, 100)}} for _ in range(num_media)]
    for media_id, media_spec, media_update in zip(media_ids, media_specs, media_updates):
        media_spec["attributes"]["test_int"] = media_update["attributes"]["test_int"]
        tator_api.update_media(media_id, media_update)
    # tator_api.update_media_list(project, media_updates[1])

    # Update tests
    patch_changes = []
    for media_id, media_spec, create_change in zip(media_ids, media_specs, create_changes):
        changes = tator_api.get_change_log_list(
            project=project, entity_id=media_id, entity_type="media"
        )

        # Assert two changes returned
        assert len(changes) == 2
        for new_change_log in changes:
            if new_change_log.id != create_change.id:
                break
        else:
            assert False, "No new change log detected"

        # Assert all old values are from creation
        compare_change_logs(new_change_log, create_change)

        # Assert all new values match initial creation values
        for change in new_change_log.description_of_change.new:
            if change.name == "_id":
                assert change.value == media_id
            elif change.name in ["_type", "_uid", "_gid", "_name", "_md5", "_section"]:
                assert change.value == media_spec[change.name.replace("_", "")]
            elif change.name.startswith("test_"):
                assert change.value == media_spec["attributes"][change.name]

        patch_changes.append(new_change_log)

    media_update = {"attributes": {"test_int": random.randint(0, 100)}}
    for media_spec in media_specs:
        media_spec["attributes"]["test_int"] = media_update["attributes"]["test_int"]
    tator_api.update_media_list(project, media_update, type=attribute_video_type)

    # Bulk update tests
    bulk_patch_changes = []
    for media_id, media_spec, create_change, patch_change in zip(
        media_ids, media_specs, create_changes, patch_changes
    ):
        changes = tator_api.get_change_log_list(
            project=project, entity_id=media_id, entity_type="media"
        )

        # Assert three changes returned
        assert len(changes) == 3
        for new_change_log in changes:
            if new_change_log.id not in [patch_change.id, create_change.id]:
                break
        else:
            assert False, "No new change log detected"

        # Assert all new values match initial creation values
        for change in new_change_log.description_of_change.new:
            if change.name == "_id":
                assert change.value == media_id
            elif change.name in ["_type", "_uid", "_gid", "_name", "_md5", "_section"]:
                assert change.value == media_spec[change.name.replace("_", "")]
            elif change.name.startswith("test_"):
                assert change.value == media_spec["attributes"][change.name]

        bulk_patch_changes.append(new_change_log)

    tator_api.delete_media(media_ids[0])
    tator_api.delete_media_list(project, media_id=media_ids[1:])

    # Deletion tests
    for media_id, media_spec, create_change, patch_change, bulk_change in zip(
        media_ids, media_specs, create_changes, patch_changes, bulk_patch_changes
    ):
        changes = tator_api.get_change_log_list(
            project=project, entity_id=media_id, entity_type="media"
        )

        # Assert four changes returned
        assert len(changes) == 4
        for new_change_log in changes:
            if new_change_log.id not in [
                patch_change.id,
                create_change.id,
                bulk_change.id,
            ]:
                break
        else:
            assert False, "No deletion change log detected"

        # Assert all old values are from update
        compare_change_logs(new_change_log, bulk_change)

        # Assert all new values are None for deletion
        for change in new_change_log.description_of_change.new:
            assert change.value == True


def test_leaf_type_change_log(host, token, project, leaf_type):
    tator_api = tator.get_api(host, token)

    num_leaves = 3
    leaf_specs = [random_leaf(project, leaf_type, None, True) for _ in range(num_leaves)]
    response = tator_api.create_leaf_list(project=project, body=leaf_specs)
    leaf_ids = response.id

    assert len(leaf_ids) == len(leaf_specs)

    # Creation tests
    create_changes = []
    for leaf_id, leaf_spec in zip(leaf_ids, leaf_specs):
        changes = tator_api.get_change_log_list(
            project=project, entity_id=leaf_id, entity_type="leaf"
        )

        # Assert one change returned
        assert len(changes) == 1
        changes = changes[0]

        # Assert all old values are None for creation
        for change in changes.description_of_change.old:
            assert change.value == None

        # Assert all new values match initial creation values
        for change in changes.description_of_change.new:
            if change.name == "_id":
                assert change.value == leaf_id
            elif change.name in ["_project", "_name", "_type"]:
                assert change.value == leaf_spec[change.name.replace("_", "")]
            elif change.name.startswith("test_"):
                assert change.value == leaf_spec["attributes"][change.name]
        create_changes.append(changes)

    patch_leaves = [random_leaf(project, leaf_type) for _ in range(num_leaves)]
    for leaf_id, patch_leaf in zip(leaf_ids, patch_leaves):
        tator_api.update_leaf(leaf_id, leaf_update=patch_leaf)

    # Update tests
    patch_changes = []
    for leaf_id, leaf, old_change_log in zip(leaf_ids, patch_leaves, create_changes):
        changes = tator_api.get_change_log_list(
            project=project, entity_id=leaf_id, entity_type="leaf"
        )

        # Assert two changes returned
        assert len(changes) == 2
        for new_change_log in changes:
            if new_change_log.id != old_change_log.id:
                break
        else:
            assert False, "No new change log detected"

        # Assert all old values are from creation
        compare_change_logs(new_change_log, old_change_log)

        # Assert all new values match initial creation values
        for change in new_change_log.description_of_change.new:
            if change.name == "_id":
                assert change.value == leaf_id
            elif change.name in ["_project", "_name", "_type"]:
                assert change.value == leaf[change.name.replace("_", "")]
            elif change.name.startswith("test_"):
                assert change.value == leaf["attributes"][change.name]
        patch_changes.append(new_change_log)

    bulk_patch = random_leaf(project, leaf_type)
    bulk_patch_spec = {"attributes": bulk_patch["attributes"]}
    tator_api.update_leaf_list(project, bulk_patch_spec)

    # Update tests
    bulk_changes = []
    for leaf_id, leaf, create_change, patch_change in zip(
        leaf_ids, patch_leaves, create_changes, patch_changes
    ):
        changes = tator_api.get_change_log_list(
            project=project, entity_id=leaf_id, entity_type="leaf"
        )

        # Assert two changes returned
        assert len(changes) == 3
        for new_change_log in changes:
            if new_change_log.id not in [patch_change.id, create_change.id]:
                break
        else:
            assert False, "No new change log detected"

        # Assert all new values match initial creation values
        for change in new_change_log.description_of_change.new:
            if change.name == "_id":
                assert change.value == leaf_id
            elif change.name in ["_project", "_name", "_type"]:
                assert change.value == bulk_patch[change.name.replace("_", "")]
            elif change.name.startswith("test_"):
                assert change.value == bulk_patch["attributes"][change.name]
        bulk_changes.append(new_change_log)

    # Clean up
    tator_api.delete_leaf(leaf_ids[0])
    tator_api.delete_leaf_list(project, leaf_id=leaf_ids[1:])

    # Deletion tests
    for leaf_id, bulk_change, patch_change, create_change in zip(
        leaf_ids, bulk_changes, patch_changes, create_changes
    ):
        changes = tator_api.get_change_log_list(
            project=project, entity_id=leaf_id, entity_type="leaf"
        )

        # Assert three changes returned
        assert len(changes) == 4
        for new_change_log in changes:
            if new_change_log.id not in [
                patch_change.id,
                create_change.id,
                bulk_change.id,
            ]:
                break
        else:
            assert False, "No deletion change log detected"

        # Assert all old values are from update
        compare_change_logs(new_change_log, bulk_change)

        # Assert all new values are None for deletion
        for change in new_change_log.description_of_change.new:
            assert change.value == True


def test_change_log_util(host, token, project, video_type):
    # Tests `tator.util.find_single_change`
    tator_api = tator.get_api(host, token)

    # Define media spec.
    num_media = 3
    fname = "MediaChangeLogTest.mp4"
    test_int = 42
    media_spec = {
        "type": video_type,
        "uid": str(uuid.uuid1()),
        "gid": str(uuid.uuid1()),
        "name": fname,
        "md5": str(uuid.uuid1())[:32],
        "section": "Test media change log",
        "attributes": {"test_int": test_int},
    }

    # Create the media.
    media_id = tator_api.create_media_list(project=project, body=media_spec).id

    # Look for change that shouldn't be there
    found_change = tator.util.find_single_change(
        tator_api, project, media_id, "test_int", new_value=test_int - 1
    )
    assert found_change is None

    # Get all changes for comparison
    changes = tator_api.get_change_log_list(
        project=project, entity_id=media_id, entity_type="media"
    )
    assert len(changes) == 1

    # Look for change that should be there
    found_change = tator.util.find_single_change(
        tator_api, project, media_id, "test_int", new_value=test_int
    )
    assert found_change == changes[0]

    # Change `test_int` value (tests attribute change)
    new_test_int = 54
    tator_api.update_media(media_id, {"attributes": {"test_int": new_test_int}})

    # Get all changes for comparison
    changes = tator_api.get_change_log_list(
        project=project, entity_id=media_id, entity_type="media"
    )
    assert len(changes) == 2

    # Look for change that should be there
    found_change = tator.util.find_single_change(
        tator_api, project, media_id, "test_int", old_value=test_int, new_value=new_test_int
    )
    assert found_change is not None
    assert found_change == changes[-1]

    # Look for change that shouldn't be there
    found_change = tator.util.find_single_change(
        tator_api, project, media_id, "test_int", old_value=new_test_int, new_value=test_int
    )
    assert found_change is None

    # Look for change that shouldn't be there
    found_change = tator.util.find_single_change(
        tator_api, project, media_id, "test_int", old_value=new_test_int
    )
    assert found_change is None

    # Delete media object (tests media property change)
    tator_api.delete_media(media_id)

    # Get all changes for comparison
    changes = tator_api.get_change_log_list(
        project=project, entity_id=media_id, entity_type="media"
    )
    assert len(changes) == 3

    # Look for change that should be there
    found_change = tator.util.find_single_change(
        tator_api,
        project,
        media_id,
        "_deleted",
        old_value=False,
        new_value=True,
        find_last_change=True,
    )
    assert found_change is not None
    assert found_change in changes



from __future__ import annotations
import logging
from typing import Any, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    import tator.TatorApi

from tator.openapi.tator_openapi import ChangeLog, ChangeLogDescriptionOfChangeNew

logger = logging.getLogger(__name__)


def find_single_change(
    api: tator.TatorApi,
    project_id: int,
    entity_id: int,
    field_name: str,
    old_value: Optional[Any] = None,
    new_value: Optional[Any] = None,
    find_last_change: bool = False,
) -> Optional[ChangeLog]:
    """Finds the ChangeLog containing the desired field/value combination.

    Example:

    .. code-block:: python

        api = tator.get_api(host, token)
        change_log = tator.util.find_change(api, media_id, "_deleted", new_value=True)
        print(change_log)

    At least one of `old_value` and `new_value` must be specified; if both are specified, then the
    matching changelog must match both, not just one.

    The `field_name` must be prefixed with an underscore if it is not a user-defined attribute, note
    how "_deleted" corresponds to the `Media.deleted` field.

    :param api: :class:`tator.TatorApi` object.
    :type api: tator.TatorApi
    :param project_id: Unique integer identifying a project in tator.
    :type project_id: int
    :param entity_id: Unique integer identifying an entity with tracked changes, must identify a
                      leaf, localization, media, or state.
    :type entity_id: int
    :param field_name: The name of the field in which to look for the change(s).
    :type field_name: str
    :param old_value: [Optional] The desired old value to find
    :type old_value: Optional[Any]
    :param new_value: [Optional] The desired new value to find
    :type new_value: Optional[Any]
    :param find_last_change: If True, finds the last matching ChangeLog, otherwise finds the first.
    :type find_last_change: bool
    :returns: The changelog corresponding to the desired value(s) to find or `None` if no matches
              are found.
    :rtype: Optional[ChangeLog]
    """
    if old_value is None and new_value is None:
        raise ValueError(
            "Must specify at least one of the following arguments: [old_value, new_value]"
        )

    change_log_list = api.get_change_log_list(project_id, entity_id=entity_id)

    def change_log_is_a_match(cl):
        matches = True

        if old_value is not None:
            old_change = ChangeLogDescriptionOfChangeNew(name=field_name, value=old_value)
            matches = matches and old_change in cl.description_of_change.old

        if new_value is not None:
            new_change = ChangeLogDescriptionOfChangeNew(name=field_name, value=new_value)
            matches = matches and new_change in cl.description_of_change.new

        return matches

    if find_last_change:
        change_log_list = reversed(change_log_list)
    return next((cl for cl in change_log_list if change_log_is_a_match(cl)), None)



# coding: utf-8

"""
    Tator REST API

    Interface to the Tator backend.  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from ..configuration import Configuration


class JobCluster(object):
    """
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'cert': 'str',
        'effective_permission': 'int',
        'host': 'str',
        'id': 'int',
        'name': 'str',
        'organization': 'int',
        'port': 'int',
        'token': 'str'
    }

    attribute_map = {
        'cert': 'cert',
        'effective_permission': 'effective_permission',
        'host': 'host',
        'id': 'id',
        'name': 'name',
        'organization': 'organization',
        'port': 'port',
        'token': 'token'
    }

    def __init__(self, cert=None, effective_permission=None, host=None, id=None, name=None, organization=None, port=None, token=None, local_vars_configuration=None):  # noqa: E501
        """JobCluster - a model defined in OpenAPI"""
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._cert = None
        self._effective_permission = None
        self._host = None
        self._id = None
        self._name = None
        self._organization = None
        self._port = None
        self._token = None
        self.discriminator = None

        if cert is not None:
            self.cert = cert
        if effective_permission is not None:
            self.effective_permission = effective_permission
        if host is not None:
            self.host = host
        if id is not None:
            self.id = id
        if name is not None:
            self.name = name
        if organization is not None:
            self.organization = organization
        if port is not None:
            self.port = port
        if token is not None:
            self.token = token

    @property
    def cert(self):
        """
        Certificate for accessing the job cluster.

        :return: The cert of this JobCluster. 
        :rtype: str
        """
        return self._cert

    @cert.setter
    def cert(self, cert):
        """
        Certificate for accessing the job cluster.

        :param cert: The cert of this JobCluster.
        :type: str
        """

        self._cert = cert

    @property
    def effective_permission(self):
        """
        Effective permission mask for this entity.

        :return: The effective_permission of this JobCluster. 
        :rtype: int
        """
        return self._effective_permission

    @effective_permission.setter
    def effective_permission(self, effective_permission):
        """
        Effective permission mask for this entity.

        :param effective_permission: The effective_permission of this JobCluster.
        :type: int
        """

        self._effective_permission = effective_permission

    @property
    def host(self):
        """
        Hostname where the cluster can be accessed.

        :return: The host of this JobCluster. 
        :rtype: str
        """
        return self._host

    @host.setter
    def host(self, host):
        """
        Hostname where the cluster can be accessed.

        :param host: The host of this JobCluster.
        :type: str
        """

        self._host = host

    @property
    def id(self):
        """
        Unique integer identifying the job cluster.

        :return: The id of this JobCluster. 
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """
        Unique integer identifying the job cluster.

        :param id: The id of this JobCluster.
        :type: int
        """

        self._id = id

    @property
    def name(self):
        """
        Unique name of the job cluster.

        :return: The name of this JobCluster. 
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Unique name of the job cluster.

        :param name: The name of this JobCluster.
        :type: str
        """

        self._name = name

    @property
    def organization(self):
        """
        Unique integer identifying the organization associated with the job cluster.

        :return: The organization of this JobCluster. 
        :rtype: int
        """
        return self._organization

    @organization.setter
    def organization(self, organization):
        """
        Unique integer identifying the organization associated with the job cluster.

        :param organization: The organization of this JobCluster.
        :type: int
        """

        self._organization = organization

    @property
    def port(self):
        """
        Port where the cluster can be accessed.

        :return: The port of this JobCluster. 
        :rtype: int
        """
        return self._port

    @port.setter
    def port(self, port):
        """
        Port where the cluster can be accessed.

        :param port: The port of this JobCluster.
        :type: int
        """

        self._port = port

    @property
    def token(self):
        """
        Token for accessing the job cluster.

        :return: The token of this JobCluster. 
        :rtype: str
        """
        return self._token

    @token.setter
    def token(self, token):
        """
        Token for accessing the job cluster.

        :param token: The token of this JobCluster.
        :type: str
        """

        self._token = token

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, JobCluster):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, JobCluster):
            return True

        return self.to_dict() != other.to_dict()



import tator

def get_images(file_path, media_or_state, num_images=None, width=None, height=None):
    """ Loads image saved by :meth:`tator.api.get_frame`

    Can also be used with :meth:`tator.api.get_state_graphic`.

    :returns: A list of :class:`PIL.Image.Image`

    :param file_path: Path to image file.
    :param media_or_state: :class:`tator.Media` or :class:`tator.State` object.
    :param num_images: [Optional] The number of images to unwrap if this
        is an image saved by :meth:`tator.TatorApi.get_frame`, or if the `offset`
        or `length` parameters were used in :meth:`tator.TatorApi.get_state_graphic`.
    :param width: [Optional] Width of ROI if this is an image saved by
        :meth:`tator.TatorApi.get_frame`.
    :param height: [Optional] Height of ROI if this is an image saved by
        :meth:`tator.TatorApi.get_frame`.
    """
    try:
        from PIL import Image
    except:
        raise RuntimeError("Utility get_images requires pillow to be installed: `pip install pillow`")
    # Read in raw image.
    img = Image.open(file_path)

    # Get tile width/height.
    if isinstance(media_or_state, tator.models.State):
        if num_images:
            num_localizations = num_images
        else:
            num_localizations = len(media_or_state.localizations)
        width = int(img.width / num_localizations)
        height = img.height
    elif isinstance(media_or_state, tator.models.Media):
        streaming = None
        if media_or_state.media_files:
            media_defs = media_or_state.media_files.streaming
            for media_def in media_defs:
                if streaming is None:
                    streaming = media_def
                elif media_def.resolution[0] > streaming.resolution[0]:
                    streaming = media_def
        if width is None:
            if streaming:
                width = streaming.resolution[1]
            else:
                width = media_or_state.width
        if height is None:
            if streaming:
                height = streaming.resolution[0]
            else:
                height = media_or_state.height

    # Make list of crops.
    images = []
    for top in range(0, img.height, height):
        for left in range(0, img.width, width):
            image = img.crop((left, top, left+width, top+height))
            images.append(image)
    if num_images:
        images = images[:num_images]
    return images



#!/usr/bin/env python

import argparse
import os
import shutil
import sys
from uuid import uuid1
import logging
import json
import glob
import re

from progressbar import progressbar
import requests
from tator.openapi.tator_openapi.models import MessageResponse

from ..util import md5sum
from ..util.get_api import get_api

from .create_media import create_media
from .determine_transcode import determine_transcode, update_media
from .transcode import convert_streaming
from .transcode import convert_archival
from .transcode import convert_audio
from .delete_media import delete_media
from .make_thumbnails import make_thumbnail_image, make_thumbnail_gif

if os.getenv("DD_LOGS_INJECTION"):
    import ddtrace.auto
    FORMAT = ('%(asctime)s %(levelname)s [%(name)s] [%(filename)s:%(lineno)d] '
              '[dd.service=%(dd.service)s dd.env=%(dd.env)s dd.version=%(dd.version)s dd.trace_id=%(dd.trace_id)s dd.span_id=%(dd.span_id)s] '
              '- %(message)s')
else:
    FORMAT = ('%(asctime)s %(levelname)s [%(name)s] [%(filename)s:%(lineno)d] '
              '- %(message)s')
logging.basicConfig(format=FORMAT)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

def parse_args():
    parser = argparse.ArgumentParser(description='Full transcode pipeline on a directory of files.')
    parser.add_argument('path', nargs='?', type=str,
                        help='Path to directory containing video files, or a video file, or text file with list of video files to be concatenated. '
                             'Ignored if --url is given.')
    parser.add_argument('--extension', type=str, default='mp4', help='File extension to upload. '
                                                                     'Ignored if path is a file.')
    parser.add_argument('--work_dir', type=str, help='Path to working directory. If not given, '
                                                     'the path containing videos will be used.')
    parser.add_argument('--url', type=str, help='URL where original file is hosted.')
    parser.add_argument('--host', type=str, default='https://cloud.tator.io', help='Host URL.')
    parser.add_argument('--token', type=str, help='REST API token.')
    parser.add_argument('--project', type=int, help='Unique integer specifying project ID.')
    parser.add_argument('--type', type=int, help='Unique integer specifying a media type.')
    parser.add_argument('--name', type=str, help='Name of the media.')
    parser.add_argument('--section', type=str, help='Media section name.')
    parser.add_argument('--section_id', type=int, help='Media section ID. If given `--section` is ignored.')
    parser.add_argument('--attributes', type=str, help="Attributes for media as a JSON string.")
    parser.add_argument('--media_id', type=int, help="Existing media ID, if applicable",
                        default=-1)
    parser.add_argument('--gid', type=str, help="Upload group ID.")
    parser.add_argument('--uid', type=str, help="Upload unique ID.")
    parser.add_argument('--group_to', type=int, default=1080,
                         help='Vertical resolutions below this will be transcoded with '
                              'multi-headed ffmpeg.')
    parser.add_argument('--email_spec', type=str, help="Optional email spec as a JSON string. The email will be sent upon completion of the transcode (see Email rest endpoint docs).")
    parser.add_argument('--cleanup', action='store_true',
                        help="Deletes working files after each file is transcoded and "
                             "uploaded.")
    parser.add_argument('--hwaccel', action='store_true', help="Use hardware acceleration.")
    parser.add_argument('--force_fps', type=float, default=-1, help='Force a specific fps for the video.')
    parser.add_argument(
        "--inhibit-upload", action="store_true", help="Do not upload the transcoded files to Tator."
    )
    parser.add_argument('--filter_complex', type=str, help='Allows (optionally templated) override of filter_complex argument to ffmpeg for streaming files. Substitutes {fps}, {width}, {height}, {hw_upload}, and {ridx} if those patterns are present, where width and height correspond to the output resolutions.')
    parser.add_argument('--bucket_id', type=int, help='Bucket ID if not default.')
    return parser.parse_args()

def get_file_paths(path, base):
    paths = {
        'original': path,
        'transcoded': base + '_transcoded',
        'thumbnail': base + '_thumbnail.jpg',
        'thumbnail_gif': base + '_thumbnail_gif.gif',
    }
    return paths

def compare_workloads(workloads_list):
    """
    Compare the parameters of workloads across multiple lists to ensure consistency.
    
    :param workloads_list: List of workload lists to compare.
    :return: True if all workloads are consistent, False otherwise.
    """
    if not workloads_list:
        return True  # No workloads to compare, considered consistent

    # Extract the first workload list as the reference
    reference_workloads = workloads_list[0]

    # Define a helper function to extract relevant parameters from a workload
    def extract_params(workload):
        return (
            workload['category'],
            workload['raw_height'],
            workload['raw_width'],
            tuple(workload['configs'])
        )

    # Extract parameters from the reference workloads
    reference_params = [extract_params(workload) for workload in reference_workloads]

    # Compare parameters with other workload lists
    for workloads in workloads_list[1:]:
        current_params = [extract_params(workload) for workload in workloads]
        if current_params != reference_params:
            return False  # Inconsistent parameters found

    return True  # All parameters are consistent

def transcode_single(path, args, gid):
    """Transcodes a single file.
    """
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)

    # If a URL is given and path doesn't exist, download the file to path.
    if args.url:
        path = os.path.join(args.work_dir, args.name)
        response = requests.get(args.url, stream=True)
        response.raise_for_status()
        with open(path, "wb") as fp:
            for chunk in response.iter_content(chunk_size=10485760):  # 10 MiB
                if chunk:
                    fp.write(chunk)
    elif path is None:
        raise ValueError(f"Must provide one of --url or path!")

    fnames = None
    if isinstance(path, list):
        if args.name is None:
            raise ValueError(f"Must provide --name when path is a list!")
        if args.work_dir:
            base = os.path.join(args.work_dir, args.name)
        else:
            base = args.name
        paths = get_file_paths(path[0], base)

    elif os.path.splitext(path)[-1] == '.json':
        # This transcode will concat multiple files
        with open(path,'r') as fp:
            fnames = json.load(fp)
        assert args.name is not None, "args.name must be provided"

        # Get file paths.
        if args.work_dir:
            base = os.path.join(args.work_dir, os.path.splitext(os.path.basename(args.name))[0])
        else:
            base, _ = os.path.splitext(args.name)

        # Use the first input file
        paths = get_file_paths(fnames[0], base)

    else:
        # Get file paths.
        if args.work_dir:
            base = os.path.join(args.work_dir, os.path.splitext(os.path.basename(path))[0])
        else:
            base, _ = os.path.splitext(path)
        paths = get_file_paths(path, base)

        # Get md5 for the file.
    md5 = md5sum(paths['original'])

    # Get base filename.
    if args.name:
        name = args.name
    else:
        name = os.path.basename(paths['original'])

    # Create a uid if not set.
    if args.uid is None:
        uid = str(uuid1())
    else:
        uid = args.uid

    # Create the media object.
    if args.media_id == -1:
        kwargs = {}
        if args.section_id:
            kwargs['section_id'] = args.section_id
            args.section = None
        media_id = create_media(args.host, args.token, args.project, args.type, args.section, name, md5, gid, uid, args.attributes, args.url, **kwargs)
    else:
        media_id = args.media_id

    try:
        # Make thumbnails.
        make_thumbnail_image(
            args.host,
            args.token,
            media_id,
            paths["original"],
            paths["thumbnail"],
            inhibit_upload=args.inhibit_upload,
            bucket_id=args.bucket_id,
        )

        if fnames:
            workloads_list = []
            for fname in fnames:
                # Determine transcodes that need to be done.
                workloads = determine_transcode(
                    args.host, args.token, args.type, media_id, fname, group_to=args.group_to
                )
                workloads_list.append(workloads)
                if compare_workloads(workloads_list):
                    # Use the first workload, but update path
                    workloads = workloads_list[0]
                    for workload in workloads:
                        workload['path'] = fnames
        else:
            # Determine transcodes that need to be done.
            workloads = determine_transcode(
                args.host, args.token, args.type, media_id, path, group_to=args.group_to
            )

        # Transcode the video file.
        for workload in workloads:
            category = workload['category']
            del workload['category']
            del workload['id']
            if category == 'streaming':
                convert_streaming(
                    **workload,
                    host=args.host,
                    token=args.token,
                    media=media_id,
                    outpath=paths["transcoded"],
                    hwaccel=args.hwaccel,
                    force_fps=args.force_fps,
                    inhibit_upload=args.inhibit_upload,
                    filter_complex=getattr(args, "filter_complex", None),
                    bucket_id=args.bucket_id,
                )
            elif category == 'archival':
                del workload['configs']
                convert_archival(
                    **workload,
                    host=args.host,
                    token=args.token,
                    media=media_id,
                    outpath=paths["transcoded"],
                    hwaccel=args.hwaccel,
                    inhibit_upload=args.inhibit_upload,
                    bucket_id=args.bucket_id,
                )
            elif category == 'audio':
                del workload['configs']
                del workload['raw_width']
                del workload['raw_height']
                convert_audio(
                    **workload,
                    host=args.host,
                    token=args.token,
                    media=media_id,
                    outpath=paths["transcoded"],
                    inhibit_upload=args.inhibit_upload,
                    bucket_id=args.bucket_id,
                )

        # Get lowest resolution output for making the gif
        pattern = re.compile(r'^\d+\.mp4$')
        outputs = [os.path.basename(f) for f in glob.glob(os.path.join(paths['transcoded'], '*.mp4'))]
        outputs = list(filter(pattern.match, outputs))
        resolutions = [int(os.path.splitext(f)[0]) for f in outputs]
        if any([res > 256 for res in resolutions]):
            resolutions = [res for res in resolutions if res > 256]

        if len(resolutions) == 0:
            logger.info(f"No transcodes of streaming files was performed, skipping gif creation and media update.")
            return

        # Files are resolution height names, sort by lowest
        input_res = min(resolutions)

        # Make gif thumbnail
        make_thumbnail_gif(
            args.host,
            args.token,
            media_id,
            os.path.join(paths["transcoded"], f"{input_res}.mp4"),
            paths["thumbnail_gif"],
            inhibit_upload=args.inhibit_upload,
            bucket_id=args.bucket_id,
        )

        # Patch the media with the concatenated file
        max_res = max(resolutions)
        update_media(args.host, args.token, args.type, media_id, os.path.join(paths['transcoded'], f"{max_res}.mp4"))

    except Exception as exc:
        logging.error("Encountered exception!", exc_info=True)
        if args.media_id == -1:
            delete_media(args.host, args.token, media_id)
        raise RuntimeError(f"Transcode of file {path} failed!") from exc

    # Clean up after the transcode is finished (if enabled).
    if args.cleanup:
        for key in ['transcoded', 'thumbnail', 'thumbnail_gif']:
            path = paths[key]
            if os.path.isdir(path):
                shutil.rmtree(path)
            else:
                os.remove(path)

    # Send an email.
    if isinstance(args.email_spec, str):
        try:
            email_spec = json.loads(args.email_spec)
        except json.JSONDecodeError:
            logger.warning(
                "Could not decode email_spec string '%s'", args.email_spec, exc_info=True
            )
            email_spec = None
    else:
        email_spec = args.email_spec

    if email_spec:
        api = get_api(args.host, args.token)
        response = api.send_email(args.project, email_spec=email_spec)
        if isinstance(response, MessageResponse):
            logger.info(response.message)

def transcode_main(args):
    if args.gid is None:
        gid = str(uuid1())
    else:
        gid = args.gid
    if args.path and not isinstance(args.path, list) and os.path.isdir(args.path):
        file_list = []
        for root, _, files in os.walk(args.path):
            for fname in files:
                if fname.endswith(args.extension):
                    path = os.path.join(root, fname)
                    file_list.append(path)
        for path in progressbar(file_list):
            transcode_single(path, args, gid)
    else:
        transcode_single(args.path, args, gid)

if __name__ == '__main__':
    args = parse_args()
    transcode_main(args)



import time
import pytest
import tator

def test_algo_template(host, token, organization, project, video):
    api = tator.get_api(host, token)
    response = api.create_hosted_template(organization, {
        "name": "echo",
        "url": "https://tator-ci.s3.us-east-1.amazonaws.com/echo.yaml",
        "headers": [],
        "tparams": [{"name": "message", "value": "Hello from the tator-py tests"}]
    })
    
    response = api.register_algorithm(project, {
        "name": "echo",
        "user": api.whoami().id,
        "files_per_job": 10000,
        "template": response.id,
    })
    print(response.message)

    response = api.create_job_list(project, {
        "algorithm_name": "echo",
        "media_ids": [video],
    })
    job_id = response.id[0]

    while True:
        print("Waiting for job to complete...")
        time.sleep(5)
        response = api.get_job(job_id)
        if response.status == 'Succeeded':
            break
        elif response.status == 'Failed':
            raise RuntimeError("Workflow template failed!")
          
def test_applet_template(host, token, organization, project):
    api = tator.get_api(host, token)
    response = api.create_hosted_template(organization, {
        "name": "echo",
        "url": "https://tator-ci.s3.us-east-1.amazonaws.com/echo.yaml",
        "headers": [],
        "tparams": [{"name": "message", "value": "Hello from the tator-py tests"}]
    })

    response = api.register_applet(project, {
        "name": "echo",
        "description": "test for templated applets",
        "template": response.id,
    })




# coding: utf-8

"""
    Tator REST API

    Interface to the Tator backend.  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from ..configuration import Configuration


class UploadCompletionSpec(object):
    """
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'bucket_id': 'int',
        'key': 'str',
        'parts': 'list[UploadPart]',
        'upload_id': 'str'
    }

    attribute_map = {
        'bucket_id': 'bucket_id',
        'key': 'key',
        'parts': 'parts',
        'upload_id': 'upload_id'
    }

    def __init__(self, bucket_id=None, key=None, parts=None, upload_id=None, local_vars_configuration=None):  # noqa: E501
        """UploadCompletionSpec - a model defined in OpenAPI"""
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._bucket_id = None
        self._key = None
        self._parts = None
        self._upload_id = None
        self.discriminator = None

        self.bucket_id = bucket_id
        if key is not None:
            self.key = key
        if parts is not None:
            self.parts = parts
        if upload_id is not None:
            self.upload_id = upload_id

    @property
    def bucket_id(self):
        """
        Bucket ID to use for upload. If not provided, the default. User must have permission to write to the bucket.

        :return: The bucket_id of this UploadCompletionSpec. 
        :rtype: int
        """
        return self._bucket_id

    @bucket_id.setter
    def bucket_id(self, bucket_id):
        """
        Bucket ID to use for upload. If not provided, the default. User must have permission to write to the bucket.

        :param bucket_id: The bucket_id of this UploadCompletionSpec.
        :type: int
        """
        if (self.local_vars_configuration.client_side_validation and
                bucket_id is not None and bucket_id < 1):  # noqa: E501
            raise ValueError("Invalid value for `bucket_id`, must be a value greater than or equal to `1`")  # noqa: E501

        self._bucket_id = bucket_id

    @property
    def key(self):
        """
        An object key that can be supplied to the `Transcode` or `Media` endpoint after the file has been uploaded.

        :return: The key of this UploadCompletionSpec. 
        :rtype: str
        """
        return self._key

    @key.setter
    def key(self, key):
        """
        An object key that can be supplied to the `Transcode` or `Media` endpoint after the file has been uploaded.

        :param key: The key of this UploadCompletionSpec.
        :type: str
        """

        self._key = key

    @property
    def parts(self):
        """
        List of objects containing part number for each PUT request and ETag header from each response.

        :return: The parts of this UploadCompletionSpec. 
        :rtype: list[UploadPart]
        """
        return self._parts

    @parts.setter
    def parts(self, parts):
        """
        List of objects containing part number for each PUT request and ETag header from each response.

        :param parts: The parts of this UploadCompletionSpec.
        :type: list[UploadPart]
        """

        self._parts = parts

    @property
    def upload_id(self):
        """
        An upload ID.

        :return: The upload_id of this UploadCompletionSpec. 
        :rtype: str
        """
        return self._upload_id

    @upload_id.setter
    def upload_id(self, upload_id):
        """
        An upload ID.

        :param upload_id: The upload_id of this UploadCompletionSpec.
        :type: str
        """

        self._upload_id = upload_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, UploadCompletionSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, UploadCompletionSpec):
            return True

        return self.to_dict() != other.to_dict()



def clone_version(src_api, src_version_id, dest_project, version_mapping={}, dest_api=None):
    """ Clone version.

    This can be used to clone versions from one project to another or from one
    host to another.

    Example for different host:

    .. code-block:: python

        src_api = tator.get_api(host, token)
        dest_api = tator.get_api(other_host, other_token)
        src_version_id = 1
        dest_project = 1
        version_mapping = {} # Necessary if source version has base versions.
        response = tator.util.clone_version(src_api, src_version_id, dest_project,
                                            version_mapping, dest_api)
        print(response.message)

    Example for same host:

    .. code-block:: python

        api = tator.get_api(host, token)
        src_version_id = 1
        dest_project = 1
        response = tator.util.clone_version(src_api, src_version_id, dest_project)
        print(response.message)

    :param src_api: :class:`tator.TatorApi` object corresponding to source host or only
        host if this is a clone on same host.
    :param src_version_id: Unique integer identifying version to clone.
    :param dest_project: Unique integer identifying destination project.
    :param version_mapping: Dictionary mapping source version IDs to destination version IDs.
                            Necessary if source version defines `bases` field.
    :param dest_api: :class:`tator.TatorApi` object corresponding to destination host.
    :returns: Response from version creation request.
    """
    version_obj = src_api.get_version(src_version_id)
    spec = {'name': version_obj.name,
            'description': version_obj.description,
            'show_empty': version_obj.show_empty,
            'bases': []}
    for base in version_obj.bases:
        if base in version_mapping:
            spec['bases'].append(version_mapping[base])
        else:
            raise ValueError(f"Base version with ID {base} not contained in version mapping!")
    if dest_api is None:
        dest_api = src_api
    return dest_api.create_version(dest_project, version_spec=spec)



import os
import tempfile

import tator

def test_get_clip(host, token, project, video):
    tator_api = tator.get_api(host, token)
    video_obj = tator_api.get_media(video)

    frame_ranges = ['0:100']
    clip = tator_api.get_clip(video, frame_ranges=frame_ranges)
    with tempfile.TemporaryDirectory() as td:
        video_path = os.path.join(td, 'video_clip_test.mp4')
        for progress in tator.util.download_temporary_file(tator_api, clip.file, video_path):
            print(f"Video clip download progress: {progress}%")
    tator_api.delete_temporary_file(clip.file.id)




# coding: utf-8

"""
    Tator REST API

    Interface to the Tator backend.  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from ..configuration import Configuration


class Credentials(object):
    """
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'password': 'str',
        'refresh': 'bool',
        'username': 'str'
    }

    attribute_map = {
        'password': 'password',
        'refresh': 'refresh',
        'username': 'username'
    }

    def __init__(self, password=None, refresh=None, username=None, local_vars_configuration=None):  # noqa: E501
        """Credentials - a model defined in OpenAPI"""
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._password = None
        self._refresh = None
        self._username = None
        self.discriminator = None

        self.password = password
        if refresh is not None:
            self.refresh = refresh
        self.username = username

    @property
    def password(self):
        """
        Account password.

        :return: The password of this Credentials. 
        :rtype: str
        """
        return self._password

    @password.setter
    def password(self, password):
        """
        Account password.

        :param password: The password of this Credentials.
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and password is None:  # noqa: E501
            raise ValueError("Invalid value for `password`, must not be `None`")  # noqa: E501

        self._password = password

    @property
    def refresh(self):
        """
        If true, forces generation of new token.

        :return: The refresh of this Credentials. 
        :rtype: bool
        """
        return self._refresh

    @refresh.setter
    def refresh(self, refresh):
        """
        If true, forces generation of new token.

        :param refresh: The refresh of this Credentials.
        :type: bool
        """

        self._refresh = refresh

    @property
    def username(self):
        """
        Account username.

        :return: The username of this Credentials. 
        :rtype: str
        """
        return self._username

    @username.setter
    def username(self, username):
        """
        Account username.

        :param username: The username of this Credentials.
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and username is None:  # noqa: E501
            raise ValueError("Invalid value for `username`, must not be `None`")  # noqa: E501

        self._username = username

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Credentials):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, Credentials):
            return True

        return self.to_dict() != other.to_dict()



import os
import random
import datetime
import uuid
import glob
import time
import collections

import tator

def random_media(api, project, paths, image_type):
    attributes = {
        'test_bool': random.choice([False, True]),
        'test_int': random.randint(-1000, 1000),
        'test_float': random.uniform(-1000.0, 1000.0),
        'test_enum': random.choice(['a', 'b', 'c']),
        'test_string': str(uuid.uuid1()),
        'test_datetime': datetime.datetime.now().isoformat(),
        'test_geopos': [random.uniform(-180.0, 180.0), random.uniform(-90.0, 90.0)],
        'test_float_array': [random.uniform(-1.0, 1.0) for _ in range(3)],
    }
    section = 'Media State Test'
    path = random.choice(paths)
    for progress, response in tator.util.upload_media(api, image_type, path,
                                                      attributes=attributes,
                                                      section=section):
        pass
    return response.id, attributes, section

def random_state(project, collection_type, media_ids, post=False):
    attributes = {
        'test_bool': random.choice([False, True]),
        'test_int': random.randint(-1000, 1000),
        'test_float': random.uniform(-1000.0, 1000.0),
        'test_enum': random.choice(['a', 'b', 'c']),
        'test_string': str(uuid.uuid1()),
        'test_datetime': datetime.datetime.now().isoformat(),
        'test_geopos': [random.uniform(-180.0, 180.0), random.uniform(-90.0, 90.0)],
        'test_float_array': [random.uniform(-1.0, 1.0) for _ in range(3)],
    }
    out = {
        'project': project,
        'type': collection_type,
        'media_ids': list(set(random.choices(media_ids, k=5))),
        'attributes': attributes
    }

    return {**out}

def test_media_states(host, token, project, image_type, image_set, collection_type):
    api = tator.get_api(host, token)
    paths = os.listdir(image_set)
    paths = glob.glob(os.path.join(image_set, '**/*.jpg'), recursive=True)
    # Create some random media.
    print("Uploading 20 images...")
    medias = [random_media(api, project, paths, image_type) for _ in range(20)]
    media_ids = [media[0] for media in medias]
    # Create states that include random sets of media.
    print("Creating 500 collections (media states)...")
    state_specs = [random_state(project, collection_type, media_ids, post=True) for _ in range(500)]
    response = api.create_state_list(project, state_specs)
    time.sleep(5)
    for idx, state_id in enumerate(response.id):
        state_specs[idx]['id'] = state_id
    # Do a search on all states.
    expected_states = [spec for spec in state_specs if spec['attributes']['test_bool'] == False]
    all_states = api.get_state_list(project, type=collection_type, attribute=["test_bool::false"])
    assert(len(all_states) == len(expected_states))

    # Do a search on all states.
    expected_states = [spec for spec in state_specs if spec['attributes']['test_bool'] == False]
    all_states = api.get_state_list_by_id(project,{'object_search':{"attribute": 'test_bool', 'operation': 'eq', 'value': False}}, type=collection_type)
    assert(len(all_states) == len(expected_states))

    # Do a search on paginated states.
    states = []
    for start in [0, 100, 200, 300, 400]:
        got_exception = False
        try:
            this_states = api.get_state_list(project, type=collection_type, attribute=["test_bool::false"],start=start, stop=start+100)
            states.extend(this_states)
        except tator.openapi.tator_openapi.exceptions.ApiException as e:
            got_exception = True
            if e.status >= 400:
                # This is expected if we have less than page_size files left
                break
            else:
                raise e
        if start > len(all_states):
            assert(got_exception | len(this_states) == 0)
        else:
            assert(not got_exception)

    assert(len(states) == len(expected_states))
    assert(collections.Counter([state.id for state in states]) == \
           collections.Counter([state.id for state in all_states]))



from datetime import datetime
import random
from types import GeneratorType
from uuid import uuid1

import tator


def make_image(tator_api, image_type, image_file, test_string):
    attributes = {"test_string": test_string}
    for progress, response in tator.util.upload_media(
        tator_api, image_type, image_file, attributes=attributes
    ):
        print(f"Upload image progress: {progress}%")
    return response.id


def random_localization(project, box_type, image_obj, test_string):
    x = random.uniform(0.0, 1.0)
    y = random.uniform(0.0, 1.0)
    w = random.uniform(0.0, 1.0 - x)
    h = random.uniform(0.0, 1.0 - y)
    attributes = {
        "test_bool": random.choice([False, True]),
        "test_int": random.randint(-1000, 1000),
        "test_float": random.uniform(-1000.0, 1000.0),
        "test_enum": random.choice(["a", "b", "c"]),
        "test_string": test_string,
        "test_datetime": datetime.now().isoformat(),
        "test_geopos": [random.uniform(-180.0, 180.0), random.uniform(-90.0, 90.0)],
        "test_float_array": [random.uniform(-1.0, 1.0) for _ in range(3)],
    }
    out = {
        "x": x,
        "y": y,
        "width": w,
        "height": h,
        "project": project,
        "type": box_type,
        "media_id": image_obj.id,
        "frame": 0,
        "attributes": attributes,
    }

    return {**out}


def _assert_pagination(api, function_name, batch_size, total, **kwargs):
    # Get paginator for `get_localization_list`
    paginator = tator.util.get_paginator(api, function_name, batch_size=batch_size)
    page_iter = paginator.paginate(**kwargs)
    assert type(page_iter) == GeneratorType

    # Check the batch size
    all_objects = []
    expect_stop_iteration = False
    while True:
        try:
            response = next(page_iter)
        except StopIteration:
            assert expect_stop_iteration
            break
        else:
            assert expect_stop_iteration == False
            all_objects += response
            if len(response) < batch_size:
                expect_stop_iteration = True
            else:
                assert len(response) == batch_size

    assert len(all_objects) == total


def test_localization_pagination(host, token, project, image_type, image_file, box_type):
    test_string = str(uuid1())
    tator_api = tator.get_api(host, token)
    image = make_image(tator_api, image_type, image_file, test_string)
    image_obj = tator_api.get_media(image)

    batch_size = 100
    num_localizations = 2222  # Should not be a multiple of batch_size
    boxes = [
        random_localization(project, box_type, image_obj, test_string)
        for _ in range(num_localizations)
    ]
    box_ids = []
    for response in tator.util.chunked_create(
        tator_api.create_localization_list, project, body=boxes
    ):
        box_ids += response.id
    assert len(box_ids) == len(boxes)
    assert len(box_ids) % 100 != 0
    print(f"Created {len(box_ids)} boxes!")

    # Verify list is the right length
    response = tator_api.get_localization_count(project, type=box_type, media_id=[image])
    assert response == num_localizations

    # Check `get_localization_list`
    _assert_pagination(
        tator_api,
        "get_localization_list",
        batch_size,
        num_localizations,
        project=project,
        type=box_type,
        media_id=[image],
    )

    # Check `get_localization_list_by_id`
    _assert_pagination(
        tator_api,
        "get_localization_list_by_id",
        batch_size,
        num_localizations,
        project=project,
        localization_id_query={"media_ids": [image]},
    )

    tator_api.delete_media(image)


def test_big_list_pagination(host, token, project, image_type, image_file, box_type):
    test_string = str(uuid1())
    tator_api = tator.get_api(host, token)
    image = make_image(tator_api, image_type, image_file, test_string)
    image_obj = tator_api.get_media(image)

    batch_size = 1000
    num_localizations = 11111
    boxes = [
        random_localization(project, box_type, image_obj, test_string)
        for _ in range(num_localizations)
    ]
    box_ids = []
    for response in tator.util.chunked_create(
        tator_api.create_localization_list, project, body=boxes
    ):
        box_ids += response.id
    assert len(box_ids) == len(boxes)
    assert len(box_ids) % 100 != 0
    print(f"Created {len(box_ids)} boxes!")

    # Verify list is the right length
    response = tator_api.get_localization_count(project, type=box_type, media_id=[image])
    assert response == num_localizations

    # Check `get_localization_list_by_id`
    _assert_pagination(
        tator_api,
        "get_localization_list_by_id",
        batch_size,
        num_localizations,
        project=project,
        localization_id_query={"media_ids": [image]},
    )

    tator_api.delete_media(image)



#!/usr/bin/python3

# This implements a  resumable upload for large files to Tator.
# This is useful for uploading large files to Tator, as it allows for the upload to be resumed if it is interrupted.
# The protocol stores the files in parts referenced in the `archival` section of the media object.

#  To rejoin the parts, use the `tator.util.import_resumable` module

#  Example:
#  python3 resumable_upload.py --token $(cat ~/token.txt) --part-size $((512*1024)) --section "Resumable Uploads" --project 2 --type 14 ~/Videos/Dragons.mp4

import tator
from tator.util._upload_file import _upload_file
from tator.util.md5sum import md5sum

import os
import math
import tempfile
import base64
import tqdm


def continue_resumable_upload(
    api,
    media_id,
    file_path,
    part_size=100 * 1024 * 1024,
    chunk_size=10 * 1024 * 1024,
    max_workers=10,
):
    """Upload a file to a media object in a project using resumable uploads.
    This function can be used to upload large files to Tator in  resumable parts.
    :param api: Tator API object.
    :param media_id: Media ID.
    :param file_path: Path to file to upload.
    :param part_size: This is the resumable part size
    :param chunk_size: This is the chunk size for transmission

    """
    media = api.get_media(media_id)
    media_type = api.get_media_type(media.type)
    if media_type.dtype != "video":
        raise ValueError("Resumable upload is only supported for video media.")

    file_size = os.stat(file_path).st_size
    num_parts = math.ceil(file_size / part_size)

    existing_parts = []
    if media.media_files:
        for archival in media.media_files.to_dict().get("archival", []):
            raw_name, ext = os.path.splitext(archival["path"])
            # Tator adds a slug to the end of the  file comps,  remove it so  we  can  keep track of which parts are left.
            last_under = raw_name.rfind("_")
            part_name = raw_name[:last_under] + ext
            existing_parts.append(os.path.basename(part_name))

    print(f"Uploading {file_path} to media {media_id} in {num_parts} resumable parts.")

    base_name, ext = os.path.splitext(os.path.basename(file_path))
    part_size_str = base64.b64encode(str(part_size).encode()).decode()
    for part in tqdm.tqdm(range(num_parts), desc="Parts"):
        part_name = part_size_str + "_" + base_name + "_" + str(part) + ext
        if part_name in existing_parts:
            print(f"Part {part} already exists. Skipping.")
            continue
        with tempfile.TemporaryDirectory() as td:
            part_path = os.path.join(td, part_name)
            with open(file_path, "rb") as f:
                f.seek(part * part_size)
                with open(part_path, "wb") as part_file:
                    part_file.write(f.read(part_size))
            for _, upload_info in _upload_file(
                api=api,
                project=media.project,
                path=part_path,
                media_id=media_id,
                filename=part_name,
                chunk_size=chunk_size,
                max_workers=max_workers,
            ):
                pass

            media_def = {"path": upload_info.key, "codec": "fragment", "resolution": [1, 1]}
            api.create_video_file(media_id, "archival", media_def)


if __name__ == "__main__":
    """CLI frontend to resumable upload"""
    import argparse
    import sys

    parser = argparse.ArgumentParser(
        description="Upload a file to a media object in a project using resumable uploads."
    )
    parser.add_argument("--host", type=str, help="Tator host", default="https://cloud.tator.io")
    parser.add_argument("--token", type=str, help="Tator token", required=True)
    parser.add_argument("--project", type=int, help="Project ID", required=False)
    parser.add_argument("--section", type=str, help="Section Name", required=False)
    parser.add_argument("--type", type=int, help="Media Type ID", required=False)
    parser.add_argument("--media-id", type=int, help="Media ID", required=False)

    ## Optional arguments  for tuning upload
    parser.add_argument(
        "--part-size", type=int, help="Resumable part size", default=100 * 1024 * 1024
    )
    parser.add_argument(
        "--chunk-size", type=int, help="Chunk size for transmission", default=10 * 1024 * 1024
    )
    parser.add_argument(
        "--max-workers", type=int, help="Max workers for concurrent requests", default=10
    )

    parser.add_argument("file_path", type=str)
    args = parser.parse_args()
    if not any([args.media_id, bool(args.project and args.section and args.type)]):
        print("Must supply project + section + type or an existing media id")
        sys.exit(-1)

    api = tator.get_api(host=args.host, token=args.token)

    # Either use the media ID  or make a  new media
    # Of note, this will be  thumbnail less to avoid excessive thumbnail generation times
    # if the input file is roughly encoded for fast decode operations
    if args.media_id:
        media_id = args.media_id
    else:
        md5 = md5sum(args.file_path)
        response = api.create_media_list(
            args.project,
            {
                "type": args.type,
                "section": args.section,
                "md5": md5,
                "name": os.path.basename(args.file_path),
            },
        )
        media_id = response.id

    print(f"Using media ID {media_id} for upload.")
    print(f"To relaunch this upload, use the following command:")
    print(
        f"python3 {sys.argv[0]} --host {args.host} --token YOUR_TOKEN_HERE --media-id {media_id} --part-size {args.part_size} {args.file_path}"
    )
    continue_resumable_upload(
        api=api,
        media_id=media_id,
        file_path=args.file_path,
        part_size=args.part_size,
        chunk_size=args.chunk_size,
        max_workers=args.max_workers,
    )



import logging
import os
import hashlib

logger = logging.getLogger(__name__)

def make_live_stream(api, type_id, layout, name, section, feedInfo):
    """ Make a live stream tator media object.

    :param api: :class:`tator.TatorApi` object.
    :param type_id: Unique integer identifying a multi-stream media type.
    :param layout: 2 element list of integers defining layout as [rows, cols].
    :param name: Name of the file to use
    :param section: Section name. If this section does not exist it will be created.
    :param feed_info: maybeList of :class:`tator.models.LiveDefinition` object or dictionary.
    :returns: Response from media object creation.
    """

    # Fetch the stuff we need to download files
    config = api.api_client.configuration
    host = config.host
    token = config.api_key['Authorization']
    prefix = config.api_key_prefix['Authorization']

    # use a live extension, if one is not specified
    if os.path.splitext(name)[1] == '':
        name += ".live"

    if type(feedInfo) is list:
        pass
    else:
        feedInfo = [feedInfo]

    # Fetch the media type
    live_stream_type = api.get_media_type(type_id)
    project = live_stream_type.project

    assert(len(feedInfo) == layout[0]*layout[1])

    # Calculate md5sum of configuration
    digest = hashlib.md5()
    digest.update(f"{feedInfo}".encode())
    md5_of_config = digest.hexdigest()


    media_spec = {
        "name": name,
        "type": type_id,
        "section": section,
        "md5": md5_of_config,
    }
    create_resp = api.create_media_list(project, [media_spec])

    patch_object = {"live": {"layout": layout, "streams": feedInfo}}
    return api.update_media(create_resp.id[0], patch_object)




#!/usr/bin/python3

# Example:
# python3 import_resumable.py --token $(cat ~/token.txt)  16744335

import tator
import os
import tempfile
from tator.transcode.transcode import convert_archival, convert_streaming, convert_audio
from tator.transcode.make_thumbnails import make_thumbnail_image, make_thumbnail_gif

from pprint import pprint
import argparse
import requests

import cv2


def import_resumable(api, media_id, work_dir=None, explicit_config=None, streaming_configs=[]):
    """Imports a resumable upload into Tator.
    :param api: `tator.TatorApi` object.
    :param media_id: Unique integer identifying a media.
    :param work_dir: [Optional] Temporary directory for joining components.
    :param explicit_config: [Optional] Explicit configuration for  archivals (see `tator.transcode.convert_archival` for details.)
    :param streaming_configs: [Optional] List of explicit configurations for streaming (see `tator.transcode.convert_streaming` for details.)
    """
    with tempfile.TemporaryDirectory() as td:
        host = api.api_client.configuration.host
        token = api.api_client.configuration.api_key["Authorization"]
        if work_dir is None:
            work_dir = td

        media = api.get_media(media_id, presigned=60 * 60 * 24)
        if streaming_configs == []:
            media_type = api.get_media_type(media.type)
            tator_streaming_config = media_type.to_dict()["streaming_config"]
            for config in tator_streaming_config:
                streaming_configs.append(
                    f"{config['resolution']}:{config['crf']}:{config['vcodec']}:{config.get('preset','fast')}:{config.get('pixel_format','yuv420p')}"
                )

        components = [x["path"] for x in media.media_files.to_dict().get("archival", [])]
        components.sort(key=lambda x: int(os.path.splitext(x)[0].split("_")[-2]))
        _, ext = os.path.splitext(components[0])
        ext = ext.split("?")[0]

        temp_path = os.path.join(work_dir, f"output{ext}")
        with open(temp_path, "wb") as output:
            for url in components:
                with requests.get(url, stream=True) as r:
                    r.raise_for_status()
                    for chunk in r.iter_content(chunk_size=8192):
                        output.write(chunk)

        thumb_path = os.path.join(work_dir, "thumb.jpg")
        thumb_gif_path = os.path.join(work_dir, "thumb.gif")

        make_thumbnail_image(
            host=args.host,
            token=args.token,
            video_path=temp_path,
            media_id=media_id,
            thumb_path=thumb_path,
        )
        make_thumbnail_gif(
            host=args.host,
            token=args.token,
            video_path=temp_path,
            media_id=media_id,
            thumb_gif_path=thumb_gif_path,
        )
        print("Completed Thumbnail Processing.")

        #  Get video resolution
        cap = cv2.VideoCapture(temp_path)
        width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        cap.release()

        # convert video for archival,  streaming,  and audio
        outpath = os.path.join(work_dir, "archivals")
        convert_archival(
            host,
            token,
            media_id,
            temp_path,
            outpath,
            width,
            height,
            size=None,
            explicit_config=explicit_config,
        )

        convert_streaming(
            host,
            token,
            media_id,
            temp_path,
            work_dir,
            width,
            height,
            streaming_configs,
        )

        convert_audio(host, token, media_id, temp_path, work_dir)

        print("Completed Transcoding.")

    print("Now we can remove the fragments.")
    media = api.get_media(media_id)
    archival_files = media.media_files.to_dict().get("archival", [])
    fragments = [x for x in archival_files if x["codec"] == "fragment"]
    while fragments:
        # Find the frag location and iterate until all are gone.
        for frag_index, info in enumerate(archival_files):
            if info["codec"] == "fragment":
                break

        api.delete_video_file(media_id, "archival", frag_index)
        media = api.get_media(media_id)
        archival_files = media.media_files.to_dict().get("archival", [])
        fragments = [x for x in archival_files if x["codec"] == "fragment"]


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Join resumable upload components.")
    parser.add_argument("--host", type=str, default="https://cloud.tator.io")
    parser.add_argument("--token", type=str, required=True, help="Tator token.")
    parser.add_argument("--work-dir", type=str, help="Temporary directory for joining components.")
    parser.add_argument("media_id", type=int, help="Media ID to join.")
    args = parser.parse_args()

    api = tator.get_api(host=args.host, token=args.token)
    import_resumable(api, args.media_id, args.work_dir)



# coding: utf-8

"""
    Tator REST API

    Interface to the Tator backend.  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from ..configuration import Configuration


class AutocompleteService(object):
    """
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'match_any': 'bool',
        'min_level': 'int',
        'mode': 'str',
        'service_url': 'str',
        'use_common': 'bool'
    }

    attribute_map = {
        'match_any': 'match_any',
        'min_level': 'minLevel',
        'mode': 'mode',
        'service_url': 'serviceUrl',
        'use_common': 'useCommon'
    }

    def __init__(self, match_any=False, min_level=None, mode='tator', service_url=None, use_common=None, local_vars_configuration=None):  # noqa: E501
        """AutocompleteService - a model defined in OpenAPI"""
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._match_any = None
        self._min_level = None
        self._mode = None
        self._service_url = None
        self._use_common = None
        self.discriminator = None

        if match_any is not None:
            self.match_any = match_any
        if min_level is not None:
            self.min_level = min_level
        if mode is not None:
            self.mode = mode
        if service_url is not None:
            self.service_url = service_url
        if use_common is not None:
            self.use_common = use_common

    @property
    def match_any(self):
        """
        If true, autocomplete will find leaves with any part of name matching the query. Otherwise only leaf names that start with the query are returned.

        :return: The match_any of this AutocompleteService. 
        :rtype: bool
        """
        return self._match_any

    @match_any.setter
    def match_any(self, match_any):
        """
        If true, autocomplete will find leaves with any part of name matching the query. Otherwise only leaf names that start with the query are returned.

        :param match_any: The match_any of this AutocompleteService.
        :type: bool
        """

        self._match_any = match_any

    @property
    def min_level(self):
        """
        If using WoRMS, set this to the minimum returnable taxonomic levelSee https://www.marinespecies.org/rest/AphiaTaxonRanksByID/-1?AphiaID=2 for the levels, Note: 220 is species

        :return: The min_level of this AutocompleteService. 
        :rtype: int
        """
        return self._min_level

    @min_level.setter
    def min_level(self, min_level):
        """
        If using WoRMS, set this to the minimum returnable taxonomic levelSee https://www.marinespecies.org/rest/AphiaTaxonRanksByID/-1?AphiaID=2 for the levels, Note: 220 is species

        :param min_level: The min_level of this AutocompleteService.
        :type: int
        """

        self._min_level = min_level

    @property
    def mode(self):
        """
        Change flavor of autocomplete to use built-in WoRMs support. For information on WoRMs see https://www.marinespecies.org/rest/

        :return: The mode of this AutocompleteService. 
        :rtype: str
        """
        return self._mode

    @mode.setter
    def mode(self, mode):
        """
        Change flavor of autocomplete to use built-in WoRMs support. For information on WoRMs see https://www.marinespecies.org/rest/

        :param mode: The mode of this AutocompleteService.
        :type: str
        """
        allowed_values = ["tator", "worms"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and mode not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `mode` ({0}), must be one of {1}"  # noqa: E501
                .format(mode, allowed_values)
            )

        self._mode = mode

    @property
    def service_url(self):
        """
        URL of the autocomplete service.

        :return: The service_url of this AutocompleteService. 
        :rtype: str
        """
        return self._service_url

    @service_url.setter
    def service_url(self, service_url):
        """
        URL of the autocomplete service.

        :param service_url: The service_url of this AutocompleteService.
        :type: str
        """

        self._service_url = service_url

    @property
    def use_common(self):
        """
        If using WoRMS, if set to true, use common names (vernacular in their API)

        :return: The use_common of this AutocompleteService. 
        :rtype: bool
        """
        return self._use_common

    @use_common.setter
    def use_common(self, use_common):
        """
        If using WoRMS, if set to true, use common names (vernacular in their API)

        :param use_common: The use_common of this AutocompleteService.
        :type: bool
        """

        self._use_common = use_common

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, AutocompleteService):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, AutocompleteService):
            return True

        return self.to_dict() != other.to_dict()



#!/usr/bin/env python3

import argparse
import os
import sys
import tempfile
import uuid

import tator

from .extractor import process_file


if __name__ == "__main__":
    """ CLI Frontend to extractor """
    parser = argparse.ArgumentParser(description="Thumbnail Extractor")
    tator.get_parser(parser)
    parser.add_argument("--mode",
                        required=True,
                        type=str,
                        choices=['state',
                                 'localization_keyframe',
                                 'localization_thumbnail',
                                 'track_thumbnail'],
                        help="""
                        state : Whole frame capture based on state's frame
                        localization_keyframe : Whole frame on localization frame
                        localization_thumbnail : Thumbnail of localization
                        track_thumbnail : Thumbnail of track (mp4)"""
                        )
    parser.add_argument("--metadata-type-id",
                        type=int,
                        required=True)
    parser.add_argument("--output-local",
                        type=str)
    parser.add_argument("--output-tator-section",
                        type=str)
    parser.add_argument("--output-type-id",
                        type=int)
    parser.add_argument("--project",
                        required=True,
                        type=int)
    parser.add_argument('--work-dir',
                        help='Optional hint to work directory',
                        type=str)
    parser.add_argument("media_ids",
                        nargs="+")

    args = parser.parse_args()

    api = tator.get_api(args.host, args.token)

    if args.mode in ['state', 'track_thumbnail']:
        print("Using States")
        meta_ep = api.get_state_list
    else:
        print("Using Localizations")
        meta_ep = api.get_localization_list

    if not any([args.output_local, args.output_tator_section]):
        print("Must supply at least one output")
        sys.exit(-1)

    if args.output_local:
        os.makedirs(args.output_local, exist_ok=True)
    upload_gid = str(uuid.uuid1())
    for media_id in args.media_ids:
        media_element = api.get_media(media_id)
        metadata = meta_ep(args.project,
                           type=args.metadata_type_id,
                           media_id=[media_id])
        metadata = [m for m in metadata if m.attributes.get('_extracted', None) is None]
        if len(metadata) == 0:
            print(f"No non-extracted metadata for '{media_element.name}'({media_id})")
            continue

        with tempfile.TemporaryDirectory(dir=args.work_dir) as td:
            print(f"Downloading {media_element.name}")
            for p in tator.download_media(api,
                                          media_element,
                                          os.path.join(td, media_element.name)):
                print(f"\r{p}", end='', flush=True)
            print('')
            print(f"Processing {media_element.name}")
            process_file(api,
                         args.project,
                         os.path.join(td, media_element.name),
                         media_element,
                         args.mode,
                         metadata,
                         args.output_local if args.output_local else td,
                         args.output_tator_section,
                         output_type_id=args.output_type_id,
                         upload_gid=upload_gid,
                         work_dir=args.work_dir)



import random
import uuid

import pytest

import tator


def random_job_cluster_spec():
    uid = str(uuid.uuid4())
    return {
        "name": f"Job Cluster {uid}",
        "host": "test-host",
        "port": random.randint(4000, 6000),
        "token": uid,
        "cert": f"{uid}.cert",
    }

def test_job_cluster_crud(host, token, organization):
    tator_api = tator.get_api(host, token)
    user = tator_api.whoami()
    tator_api.create_affiliation(organization, {"permission": "Admin", "user": user.id})

    job_cluster_spec = random_job_cluster_spec()

    # Test creation
    response = tator_api.create_job_cluster(organization, job_cluster_spec)
    assert hasattr(response, "id")
    jc_id = response.id
    response = tator_api.get_job_cluster(jc_id)
    assert response.id == jc_id
    for k, v in job_cluster_spec.items():
        assert getattr(response, k) == v

    # Test read
    jc_list = tator_api.get_job_cluster_list(organization)
    assert len(jc_list) == 1
    job_cluster = jc_list[0]
    assert job_cluster.id == jc_id
    for k, v in job_cluster_spec.items():
        assert getattr(job_cluster, k) == v

    # Test update
    job_cluster_spec = random_job_cluster_spec()
    response = tator_api.update_job_cluster(jc_id, job_cluster_spec)
    assert response.message == f"Job Cluster {jc_id} successfully updated!"

    response = tator_api.get_job_cluster(jc_id)
    assert response.id == jc_id
    for k, v in job_cluster_spec.items():
        assert getattr(response, k) == v

    # Test delete
    response = tator_api.delete_job_cluster(jc_id)
    assert response.message == "Job cluster deleted successfully!"
    with pytest.raises(tator.openapi.tator_openapi.exceptions.ApiException):
        tator_api.get_job_cluster(jc_id)



import argparse
import os


def validate_string_argument(arg: str) -> str:
    """Validates string arguments that should not be empty"""
    if arg:
        return arg
    raise RuntimeError(f"Expected a non-empty string, received {arg=}")


def get_parser(parser=None, description=None):
    """Returns an argument parser that includes host and token.

    :returns: :class:`argparse.ArgumentParser` object that includes host and token arguments.
    """
    if parser is None:
        parser = argparse.ArgumentParser(
            description=description if description else "A tator-py utility.",
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
    parser.add_argument(
        "--host",
        type=validate_string_argument,
        default=os.getenv("TATOR_HOST", "https://cloud.tator.io"),
        help=(
            "The Tator host to connect to; if no host is provided on the command line, this will "
            "fall back to the environment variable `TATOR_HOST`."
        ),
    )
    parser.add_argument(
        "--token",
        type=validate_string_argument,
        default=os.getenv("TATOR_TOKEN", ""),
        help=(
            "The authentication token to use; if no token is provided on the command line, this "
            "will fall back to the environment variable `TATOR_TOKEN`."
        ),
    )
    return parser



import random
import uuid
import datetime

import tator


def random_localization(project, box_type, video_obj, post=False):
    x = random.uniform(0.0, 1.0)
    y = random.uniform(0.0, 1.0)
    w = random.uniform(0.0, 1.0 - x)
    h = random.uniform(0.0, 1.0 - y)
    attributes = {
        "test_bool": random.choice([False, True]),
        "test_int": random.randint(-1000, 1000),
        "test_float": random.uniform(-1000.0, 1000.0),
        "test_enum": random.choice(["a", "b", "c"]),
        "test_string": str(uuid.uuid1()),
        "test_datetime": datetime.datetime.now().isoformat(),
        "test_geopos": [random.uniform(-180.0, 180.0), random.uniform(-90.0, 90.0)],
        "test_float_array": [random.uniform(-1.0, 1.0) for _ in range(3)],
    }
    out = {
        "x": x,
        "y": y,
        "width": w,
        "height": h,
        "project": project,
        "type": box_type,
        "media_id": video_obj.id,
        "frame": random.randint(0, video_obj.num_frames - 1),
        "attributes": attributes,
    }

    return {**out}


def random_state(project, state_type, video_obj, post=False):
    attributes = {
        "test_bool": random.choice([False, True]),
        "test_int": random.randint(-1000, 1000),
        "test_float": random.uniform(-1000.0, 1000.0),
        "test_enum": random.choice(["a", "b", "c"]),
        "test_string": str(uuid.uuid1()),
        "test_datetime": datetime.datetime.now().isoformat(),
        "test_geopos": [random.uniform(-180.0, 180.0), random.uniform(-90.0, 90.0)],
        "test_float_array": [random.uniform(-1.0, 1.0) for _ in range(3)],
    }
    out = {
        "project": project,
        "type": state_type,
        "media_ids": [video_obj.id],
        "frame": random.randint(0, video_obj.num_frames - 1),
        "attributes": attributes,
    }

    return {**out}


def test_append(host, token, project, video_type, video, track_type, box_type):
    tator_api = tator.get_api(host, token)
    video_obj = tator_api.get_media(video)

    # Create some boxes.
    num_localizations = random.randint(5, 10)
    boxes = [
        random_localization(project, box_type, video_obj, post=True)
        for _ in range(num_localizations)
    ]
    box_ids = []
    for response in tator.util.chunked_create(
        tator_api.create_localization_list, project, body=boxes
    ):
        box_ids += response.id
    assert len(box_ids) == len(boxes)
    print(f"Created {len(box_ids)} boxes!")

    # Create a track.
    track = random_state(project, track_type, video_obj, post=True)
    response = tator_api.create_state_list(project, track)
    assert isinstance(response, tator.models.CreateListResponse)
    print(response.message)
    state_id = response.id[0]
    state_obj = tator_api.get_state(state_id)
    elemental_id = state_obj.elemental_id
    version = state_obj.version

    # Add boxes to track one at a time.
    for box_id in box_ids:
        for _ in range(2):
            update = {"localization_ids_add": [box_id]}
            response = tator_api.update_state_by_elemental_id(
                version, elemental_id, state_update=update
            )
            assert "message" in response.to_dict()

    # Add boxes to track.
    update = {"localization_ids_add": box_ids}
    response = tator_api.update_state_by_elemental_id(version, elemental_id, state_update=update)
    assert "message" in response.to_dict()

    # Try adding boxes to track again.
    response = tator_api.update_state_by_elemental_id(version, elemental_id, state_update=update)
    assert "message" in response.to_dict()
    print(response.message)

    # Try adding boxes to track with repeated entries.
    update = {"localization_ids_add": box_ids + box_ids}
    response = tator_api.update_state_by_elemental_id(version, elemental_id, state_update=update)
    assert "message" in response.to_dict()
    print(response.message)

    # Make sure we have the expected number of added localizations.
    track = tator_api.get_state_by_elemental_id(version, elemental_id)
    assert len(track.localizations) == num_localizations



# coding: utf-8

"""
    Tator REST API

    Interface to the Tator backend.  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from ..configuration import Configuration


class StateDelete(object):
    """
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'pedantic': 'int',
        'prune': 'int'
    }

    attribute_map = {
        'pedantic': 'pedantic',
        'prune': 'prune'
    }

    def __init__(self, pedantic=0, prune=0, local_vars_configuration=None):  # noqa: E501
        """StateDelete - a model defined in OpenAPI"""
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._pedantic = None
        self._prune = None
        self.discriminator = None

        if pedantic is not None:
            self.pedantic = pedantic
        if prune is not None:
            self.prune = prune

    @property
    def pedantic(self):
        """
        Set to 1 to enforce that this is the latest mark; else push edits to end of branch.

        :return: The pedantic of this StateDelete. 
        :rtype: int
        """
        return self._pedantic

    @pedantic.setter
    def pedantic(self, pedantic):
        """
        Set to 1 to enforce that this is the latest mark; else push edits to end of branch.

        :param pedantic: The pedantic of this StateDelete.
        :type: int
        """
        if (self.local_vars_configuration.client_side_validation and
                pedantic is not None and pedantic > 1):  # noqa: E501
            raise ValueError("Invalid value for `pedantic`, must be a value less than or equal to `1`")  # noqa: E501
        if (self.local_vars_configuration.client_side_validation and
                pedantic is not None and pedantic < 0):  # noqa: E501
            raise ValueError("Invalid value for `pedantic`, must be a value greater than or equal to `0`")  # noqa: E501

        self._pedantic = pedantic

    @property
    def prune(self):
        """
        If set to 1 will purge the object from the database entirely. This removes any record, change-log, that this metadatum ever existed.

        :return: The prune of this StateDelete. 
        :rtype: int
        """
        return self._prune

    @prune.setter
    def prune(self, prune):
        """
        If set to 1 will purge the object from the database entirely. This removes any record, change-log, that this metadatum ever existed.

        :param prune: The prune of this StateDelete.
        :type: int
        """
        if (self.local_vars_configuration.client_side_validation and
                prune is not None and prune > 1):  # noqa: E501
            raise ValueError("Invalid value for `prune`, must be a value less than or equal to `1`")  # noqa: E501
        if (self.local_vars_configuration.client_side_validation and
                prune is not None and prune < 0):  # noqa: E501
            raise ValueError("Invalid value for `prune`, must be a value greater than or equal to `0`")  # noqa: E501

        self._prune = prune

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, StateDelete):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, StateDelete):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Tator REST API

    Interface to the Tator backend.  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from ..configuration import Configuration


class JobSpec(object):
    """
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'algorithm_name': 'str',
        'extra_params': 'list[AlgorithmParameter]',
        'failure_email_spec': 'JobSpecFailureEmailSpec',
        'media_ids': 'list[int]',
        'success_email_spec': 'JobSpecFailureEmailSpec'
    }

    attribute_map = {
        'algorithm_name': 'algorithm_name',
        'extra_params': 'extra_params',
        'failure_email_spec': 'failure_email_spec',
        'media_ids': 'media_ids',
        'success_email_spec': 'success_email_spec'
    }

    def __init__(self, algorithm_name=None, extra_params=None, failure_email_spec=None, media_ids=None, success_email_spec=None, local_vars_configuration=None):  # noqa: E501
        """JobSpec - a model defined in OpenAPI"""
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._algorithm_name = None
        self._extra_params = None
        self._failure_email_spec = None
        self._media_ids = None
        self._success_email_spec = None
        self.discriminator = None

        self.algorithm_name = algorithm_name
        if extra_params is not None:
            self.extra_params = extra_params
        self.failure_email_spec = failure_email_spec
        self.media_ids = media_ids
        self.success_email_spec = success_email_spec

    @property
    def algorithm_name(self):
        """
        Name of the algorithm to execute.

        :return: The algorithm_name of this JobSpec. 
        :rtype: str
        """
        return self._algorithm_name

    @algorithm_name.setter
    def algorithm_name(self, algorithm_name):
        """
        Name of the algorithm to execute.

        :param algorithm_name: The algorithm_name of this JobSpec.
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and algorithm_name is None:  # noqa: E501
            raise ValueError("Invalid value for `algorithm_name`, must not be `None`")  # noqa: E501

        self._algorithm_name = algorithm_name

    @property
    def extra_params(self):
        """
        Extra parameters to pass into the algorithm

        :return: The extra_params of this JobSpec. 
        :rtype: list[AlgorithmParameter]
        """
        return self._extra_params

    @extra_params.setter
    def extra_params(self, extra_params):
        """
        Extra parameters to pass into the algorithm

        :param extra_params: The extra_params of this JobSpec.
        :type: list[AlgorithmParameter]
        """

        self._extra_params = extra_params

    @property
    def failure_email_spec(self):
        """

        :return: The failure_email_spec of this JobSpec. 
        :rtype: JobSpecFailureEmailSpec
        """
        return self._failure_email_spec

    @failure_email_spec.setter
    def failure_email_spec(self, failure_email_spec):
        """

        :param failure_email_spec: The failure_email_spec of this JobSpec.
        :type: JobSpecFailureEmailSpec
        """

        self._failure_email_spec = failure_email_spec

    @property
    def media_ids(self):
        """
        List of media IDs.

        :return: The media_ids of this JobSpec. 
        :rtype: list[int]
        """
        return self._media_ids

    @media_ids.setter
    def media_ids(self, media_ids):
        """
        List of media IDs.

        :param media_ids: The media_ids of this JobSpec.
        :type: list[int]
        """
        if self.local_vars_configuration.client_side_validation and media_ids is None:  # noqa: E501
            raise ValueError("Invalid value for `media_ids`, must not be `None`")  # noqa: E501

        self._media_ids = media_ids

    @property
    def success_email_spec(self):
        """

        :return: The success_email_spec of this JobSpec. 
        :rtype: JobSpecFailureEmailSpec
        """
        return self._success_email_spec

    @success_email_spec.setter
    def success_email_spec(self, success_email_spec):
        """

        :param success_email_spec: The success_email_spec of this JobSpec.
        :type: JobSpecFailureEmailSpec
        """

        self._success_email_spec = success_email_spec

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, JobSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, JobSpec):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Tator REST API

    Interface to the Tator backend.  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from ..configuration import Configuration


class AttributeOperationSpec(object):
    """
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'method': 'str',
        'operations': 'list[object]',
        'attribute': 'str',
        'inverse': 'bool',
        'operation': 'str',
        'value': 'object'
    }

    attribute_map = {
        'method': 'method',
        'operations': 'operations',
        'attribute': 'attribute',
        'inverse': 'inverse',
        'operation': 'operation',
        'value': 'value'
    }

    def __init__(self, method=None, operations=None, attribute=None, inverse=False, operation=None, value=None, local_vars_configuration=None):  # noqa: E501
        """AttributeOperationSpec - a model defined in OpenAPI"""
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._method = None
        self._operations = None
        self._attribute = None
        self._inverse = None
        self._operation = None
        self._value = None
        self.discriminator = None

        if method is not None:
            self.method = method
        if operations is not None:
            self.operations = operations
        if attribute is not None:
            self.attribute = attribute
        if inverse is not None:
            self.inverse = inverse
        if operation is not None:
            self.operation = operation
        self.value = value

    @property
    def method(self):
        """
        Method to combine 

        :return: The method of this AttributeOperationSpec. 
        :rtype: str
        """
        return self._method

    @method.setter
    def method(self, method):
        """
        Method to combine 

        :param method: The method of this AttributeOperationSpec.
        :type: str
        """
        allowed_values = ["and", "or", "not"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and method not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `method` ({0}), must be one of {1}"  # noqa: E501
                .format(method, allowed_values)
            )

        self._method = method

    @property
    def operations(self):
        """
        Must be a AttributeOperationSpec!

        :return: The operations of this AttributeOperationSpec. 
        :rtype: list[object]
        """
        return self._operations

    @operations.setter
    def operations(self, operations):
        """
        Must be a AttributeOperationSpec!

        :param operations: The operations of this AttributeOperationSpec.
        :type: list[object]
        """

        self._operations = operations

    @property
    def attribute(self):
        """
        Name of the attribute or field to search on.                               Given a user defined type (Localization, State, or Media) with defined attributes                any attribute is searchable by its name. If a Localization has an attribute named 'Species'                the name to use to search on that column is 'Species'.                 Additional built-in columns are available to search on dependent on the underlying metadata               type being searched. To search a built in field, a '$' character must be used in front of the column               name from main.models.                              The following table shows common columns and to which types they apply. Any internal column                of the model may be searched in this manner.                              '$' must precede these names in search attempts. E.g. `created_by` is supplied as `$created_by`.                <table border=\"1\">     <tr>         <th>Name</th>         <th>Description</th>         <th>Localizations</th>         <th>States</th>         <th>Medias</th>         <th>Leaves</th>         <th>Files</th>     </tr>     <tr>         <td>section</td>         <td>Media section</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td></td>         <td></td>     </tr>     <tr>         <td>created_datetime</td>         <td>The time of creation for this datum</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td>X</td>     </tr>     <tr>         <td>created_by</td>         <td>The user id who created this datum</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td>X</td>     </tr>     <tr>         <td>modified_datetime</td>         <td>The last modification time</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td>X</td>     </tr>     <tr>         <td>modified_by</td>         <td>The last modification user</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td>X</td>     </tr>     <tr>         <td>name</td>         <td>The name of the element</td>         <td></td>         <td></td>         <td>X</td>         <td>X</td>         <td>X</td>     </tr>     <tr>         <td>fps</td>         <td>The frames per second</td>         <td></td>         <td></td>         <td>X</td>         <td></td>         <td></td>     </tr>     <tr>         <td>deleted</td>         <td>Whether the media is marked deleted</td>         <td></td>         <td></td>         <td>X</td>         <td>X</td>         <td>X</td>     </tr>     <tr>         <td>variant_deleted</td>         <td>Whether the metadata is marked deleted</td>         <td>X</td>         <td>X</td>         <td></td>         <td></td>         <td></td>     </tr>     <tr>         <td>archive_state</td>         <td>The current archive state of the media</td>         <td></td>         <td></td>         <td>X</td>         <td></td>         <td></td>     </tr>     <tr>         <td>x, y, u, or v</td>         <td>Geometric coordinates</td>         <td></td>         <td></td>         <td></td>         <td></td>         <td></td>     </tr>     <tr>         <td>width or height</td>         <td>Geometric sizes</td>         <td>x</td>         <td></td>         <td></td>         <td></td>         <td></td>     </tr>     <tr>         <td>incident</td>         <td>Available when doing a related search</td>         <td></td>         <td></td>         <td>X</td>         <td></td>         <td></td>     </tr>     <tr>         <td>parent</td>         <td>Foreign key to parent element</td>         <td>X</td>         <td>X</td>         <td></td>         <td></td>         <td></td>     </tr> </table> <br /><br />The following special columns are available for searching on related data. Also must be preceded with a '$'. <table border=\"1\">     <tr>         <th>Name</th>         <th>Description</th>         <th>Localizations</th>         <th>States</th>         <th>Medias</th>         <th>Leaves</th>         <th>Files</th>     </tr>     <tr>         <td>related_localizations</td>         <td>Returns media that contain localizations that match this subquery</td>         <td></td>         <td></td>         <td>X</td>         <td></td>         <td></td>     </tr>     <tr>         <td>related_states</td>         <td>Returns media that contain states that match this subquery</td>         <td></td>         <td></td>         <td>X</td>         <td></td>         <td></td>     </tr>     <tr>         <td>related_media</td>         <td>Returns metadata that is associated with media matching this subquery</td>         <td>X</td>         <td>X</td>         <td></td>         <td></td>         <td></td>     </tr>     <tr>         <td>track_membership</td>         <td>Returns localizations that are in a matching track (localization-state)</td>         <td>X</td>         <td></td>         <td></td>         <td></td>         <td></td>     </tr>     <tr>         <td>coincident_states</td>         <td>Returns metadata that is coincident with state(s) (same frame/media)</td>         <td>X</td>         <td>X</td>         <td></td>         <td></td>         <td></td>     </tr>     <tr>         <td>coincident_localizations</td>         <td>Returns metadata that is coincident with a localization(s) (same frame/media)</td>         <td>X</td>         <td>X</td>         <td></td>         <td></td>         <td></td>     </tr> </table> <b>IMPORTANT:</b> Only the 'search' operation is valid for these special columns

        :return: The attribute of this AttributeOperationSpec. 
        :rtype: str
        """
        return self._attribute

    @attribute.setter
    def attribute(self, attribute):
        """
        Name of the attribute or field to search on.                               Given a user defined type (Localization, State, or Media) with defined attributes                any attribute is searchable by its name. If a Localization has an attribute named 'Species'                the name to use to search on that column is 'Species'.                 Additional built-in columns are available to search on dependent on the underlying metadata               type being searched. To search a built in field, a '$' character must be used in front of the column               name from main.models.                              The following table shows common columns and to which types they apply. Any internal column                of the model may be searched in this manner.                              '$' must precede these names in search attempts. E.g. `created_by` is supplied as `$created_by`.                <table border=\"1\">     <tr>         <th>Name</th>         <th>Description</th>         <th>Localizations</th>         <th>States</th>         <th>Medias</th>         <th>Leaves</th>         <th>Files</th>     </tr>     <tr>         <td>section</td>         <td>Media section</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td></td>         <td></td>     </tr>     <tr>         <td>created_datetime</td>         <td>The time of creation for this datum</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td>X</td>     </tr>     <tr>         <td>created_by</td>         <td>The user id who created this datum</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td>X</td>     </tr>     <tr>         <td>modified_datetime</td>         <td>The last modification time</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td>X</td>     </tr>     <tr>         <td>modified_by</td>         <td>The last modification user</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td>X</td>         <td>X</td>     </tr>     <tr>         <td>name</td>         <td>The name of the element</td>         <td></td>         <td></td>         <td>X</td>         <td>X</td>         <td>X</td>     </tr>     <tr>         <td>fps</td>         <td>The frames per second</td>         <td></td>         <td></td>         <td>X</td>         <td></td>         <td></td>     </tr>     <tr>         <td>deleted</td>         <td>Whether the media is marked deleted</td>         <td></td>         <td></td>         <td>X</td>         <td>X</td>         <td>X</td>     </tr>     <tr>         <td>variant_deleted</td>         <td>Whether the metadata is marked deleted</td>         <td>X</td>         <td>X</td>         <td></td>         <td></td>         <td></td>     </tr>     <tr>         <td>archive_state</td>         <td>The current archive state of the media</td>         <td></td>         <td></td>         <td>X</td>         <td></td>         <td></td>     </tr>     <tr>         <td>x, y, u, or v</td>         <td>Geometric coordinates</td>         <td></td>         <td></td>         <td></td>         <td></td>         <td></td>     </tr>     <tr>         <td>width or height</td>         <td>Geometric sizes</td>         <td>x</td>         <td></td>         <td></td>         <td></td>         <td></td>     </tr>     <tr>         <td>incident</td>         <td>Available when doing a related search</td>         <td></td>         <td></td>         <td>X</td>         <td></td>         <td></td>     </tr>     <tr>         <td>parent</td>         <td>Foreign key to parent element</td>         <td>X</td>         <td>X</td>         <td></td>         <td></td>         <td></td>     </tr> </table> <br /><br />The following special columns are available for searching on related data. Also must be preceded with a '$'. <table border=\"1\">     <tr>         <th>Name</th>         <th>Description</th>         <th>Localizations</th>         <th>States</th>         <th>Medias</th>         <th>Leaves</th>         <th>Files</th>     </tr>     <tr>         <td>related_localizations</td>         <td>Returns media that contain localizations that match this subquery</td>         <td></td>         <td></td>         <td>X</td>         <td></td>         <td></td>     </tr>     <tr>         <td>related_states</td>         <td>Returns media that contain states that match this subquery</td>         <td></td>         <td></td>         <td>X</td>         <td></td>         <td></td>     </tr>     <tr>         <td>related_media</td>         <td>Returns metadata that is associated with media matching this subquery</td>         <td>X</td>         <td>X</td>         <td></td>         <td></td>         <td></td>     </tr>     <tr>         <td>track_membership</td>         <td>Returns localizations that are in a matching track (localization-state)</td>         <td>X</td>         <td></td>         <td></td>         <td></td>         <td></td>     </tr>     <tr>         <td>coincident_states</td>         <td>Returns metadata that is coincident with state(s) (same frame/media)</td>         <td>X</td>         <td>X</td>         <td></td>         <td></td>         <td></td>     </tr>     <tr>         <td>coincident_localizations</td>         <td>Returns metadata that is coincident with a localization(s) (same frame/media)</td>         <td>X</td>         <td>X</td>         <td></td>         <td></td>         <td></td>     </tr> </table> <b>IMPORTANT:</b> Only the 'search' operation is valid for these special columns

        :param attribute: The attribute of this AttributeOperationSpec.
        :type: str
        """

        self._attribute = attribute

    @property
    def inverse(self):
        """
        Whether to apply NOT to result

        :return: The inverse of this AttributeOperationSpec. 
        :rtype: bool
        """
        return self._inverse

    @inverse.setter
    def inverse(self, inverse):
        """
        Whether to apply NOT to result

        :param inverse: The inverse of this AttributeOperationSpec.
        :type: bool
        """

        self._inverse = inverse

    @property
    def operation(self):
        """
        Name of the operation to apply. Not all operations apply to all attributes.

        :return: The operation of this AttributeOperationSpec. 
        :rtype: str
        """
        return self._operation

    @operation.setter
    def operation(self, operation):
        """
        Name of the operation to apply. Not all operations apply to all attributes.

        :param operation: The operation of this AttributeOperationSpec.
        :type: str
        """
        allowed_values = ["eq", "gt", "gte", "lt", "lte", "icontains", "iendswidth", "istartswith", "isnull", "in", "date_eq", "date_gte", "date_gt", "date_lt", "date_lte", "date_range", "distance_lte", "search"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and operation not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `operation` ({0}), must be one of {1}"  # noqa: E501
                .format(operation, allowed_values)
            )

        self._operation = operation

    @property
    def value(self):
        """
        Boolean, integer, float, string, datetime, [lon, lat], float array.

        :return: The value of this AttributeOperationSpec. 
        :rtype: object
        """
        return self._value

    @value.setter
    def value(self, value):
        """
        Boolean, integer, float, string, datetime, [lon, lat], float array.

        :param value: The value of this AttributeOperationSpec.
        :type: object
        """

        self._value = value

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, AttributeOperationSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, AttributeOperationSpec):
            return True

        return self.to_dict() != other.to_dict()



from datetime import datetime
import random
from uuid import uuid1

import pytest
import tator


def make_image(tator_api, image_type, image_file, test_string):
    attributes = {"test_string": test_string}
    for progress, response in tator.util.upload_media(
        tator_api, image_type, image_file, attributes=attributes
    ):
        print(f"Upload image progress: {progress}%")
    return response.id


def random_localization(project, box_type, image_obj, test_string):
    x = random.uniform(0.0, 1.0)
    y = random.uniform(0.0, 1.0)
    w = random.uniform(0.0, 1.0 - x)
    h = random.uniform(0.0, 1.0 - y)
    attributes = {
        "test_bool": random.choice([False, True]),
        "test_int": random.randint(-1000, 1000),
        "test_float": random.uniform(-1000.0, 1000.0),
        "test_enum": random.choice(["a", "b", "c"]),
        "test_string": test_string,
        "test_datetime": datetime.now().isoformat(),
        "test_geopos": [random.uniform(-180.0, 180.0), random.uniform(-90.0, 90.0)],
        "test_float_array": [random.uniform(-1.0, 1.0) for _ in range(3)],
    }
    out = {
        "x": x,
        "y": y,
        "width": w,
        "height": h,
        "project": project,
        "type": box_type,
        "media_id": image_obj.id,
        "frame": 0,
        "attributes": attributes,
    }

    return {**out}


def test_localization_chunked_create(host, token, project, image_type, image_file, box_type):
    test_string = str(uuid1())
    tator_api = tator.get_api(host, token)
    image = make_image(tator_api, image_type, image_file, test_string)
    image_obj = tator_api.get_media(image)

    num_localizations = 1111
    boxes = [
        random_localization(project, box_type, image_obj, test_string)
        for _ in range(num_localizations)
    ]

    # Default chunk_size
    box_ids = [
        new_id
        for response in tator.util.chunked_create(
            tator_api.create_localization_list, project, body=boxes
        )
        for new_id in response.id
    ]
    assert len(box_ids) == len(boxes)

    # Small chunk_size
    box_ids = [
        new_id
        for response in tator.util.chunked_create(
            tator_api.create_localization_list, project, chunk_size=100, body=boxes
        )
        for new_id in response.id
    ]
    assert len(box_ids) == len(boxes)

    # Huge chunk_size
    box_ids = [
        new_id
        for response in tator.util.chunked_create(
            tator_api.create_localization_list, project, chunk_size=1000, body=boxes
        )
        for new_id in response.id
    ]
    assert len(box_ids) == len(boxes)

    tator_api.delete_media(image)


def helper(api, project, boxes, chunk_size=None):
    kwargs = {}
    if chunk_size:
        kwargs["chunk_size"] = chunk_size
    box_ids = []
    for response in tator.util.chunked_create(
        api.create_localization_list, project, body=boxes, **kwargs
    ):
        box_ids += response.id
        print(f"Created {len(response.id)} boxes")
    assert len(box_ids) == len(boxes)


def test_localization_chunked_create(host, token, project, image_type, image_file, box_type):
    test_string = str(uuid1())
    tator_api = tator.get_api(host, token)
    image = make_image(tator_api, image_type, image_file, test_string)
    image_obj = tator_api.get_media(image)

    num_localizations = 1111
    boxes = [
        random_localization(project, box_type, image_obj, test_string)
        for _ in range(num_localizations)
    ]

    # Default chunk_size
    print("using default chunk size")
    helper(tator_api, project, boxes)

    # Small chunk_size
    print("using small chunk size (100)")
    helper(tator_api, project, boxes, chunk_size=100)

    # Huge chunk_size
    print("using huge chunk size (1000)")
    helper(tator_api, project, boxes, chunk_size=1000)

    # Give bad func to force a failure to create
    with pytest.raises(RuntimeError) as exc:
        for _ in tator.util.chunked_create(lambda _: True, project, body=boxes):
            pass

    n_boxes = len(boxes)
    assert str(exc.value) == f"Was not able to create {n_boxes} of {n_boxes} objects"

    tator_api.delete_media(image)



# coding: utf-8

"""
    Tator REST API

    Interface to the Tator backend.  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from ..configuration import Configuration


class MediaBulkUpdate(object):
    """
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'archive_state': 'str',
        'attributes': 'dict(str, object)',
        'float_array': 'list[FloatArrayQuery]',
        'ids': 'list[int]',
        'localization_ids': 'list[int]',
        'null_attributes': 'list[str]',
        'object_search': 'AttributeOperationSpec',
        'reset_attributes': 'list[str]',
        'state_ids': 'list[int]',
        'user_elemental_id': 'str'
    }

    attribute_map = {
        'archive_state': 'archive_state',
        'attributes': 'attributes',
        'float_array': 'float_array',
        'ids': 'ids',
        'localization_ids': 'localization_ids',
        'null_attributes': 'null_attributes',
        'object_search': 'object_search',
        'reset_attributes': 'reset_attributes',
        'state_ids': 'state_ids',
        'user_elemental_id': 'user_elemental_id'
    }

    def __init__(self, archive_state=None, attributes=None, float_array=None, ids=None, localization_ids=None, null_attributes=None, object_search=None, reset_attributes=None, state_ids=None, user_elemental_id=None, local_vars_configuration=None):  # noqa: E501
        """MediaBulkUpdate - a model defined in OpenAPI"""
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._archive_state = None
        self._attributes = None
        self._float_array = None
        self._ids = None
        self._localization_ids = None
        self._null_attributes = None
        self._object_search = None
        self._reset_attributes = None
        self._state_ids = None
        self._user_elemental_id = None
        self.discriminator = None

        if archive_state is not None:
            self.archive_state = archive_state
        if attributes is not None:
            self.attributes = attributes
        if float_array is not None:
            self.float_array = float_array
        if ids is not None:
            self.ids = ids
        if localization_ids is not None:
            self.localization_ids = localization_ids
        if null_attributes is not None:
            self.null_attributes = null_attributes
        if object_search is not None:
            self.object_search = object_search
        if reset_attributes is not None:
            self.reset_attributes = reset_attributes
        if state_ids is not None:
            self.state_ids = state_ids
        if user_elemental_id is not None:
            self.user_elemental_id = user_elemental_id

    @property
    def archive_state(self):
        """
        Marks media for archival or retrieval. Media may not be set directly to `live` or `archived`, the system performs that transition for the user.

        :return: The archive_state of this MediaBulkUpdate. 
        :rtype: str
        """
        return self._archive_state

    @archive_state.setter
    def archive_state(self, archive_state):
        """
        Marks media for archival or retrieval. Media may not be set directly to `live` or `archived`, the system performs that transition for the user.

        :param archive_state: The archive_state of this MediaBulkUpdate.
        :type: str
        """
        allowed_values = ["to_archive", "to_live"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and archive_state not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `archive_state` ({0}), must be one of {1}"  # noqa: E501
                .format(archive_state, allowed_values)
            )

        self._archive_state = archive_state

    @property
    def attributes(self):
        """
        Attribute values to bulk update an entity list.

        :return: The attributes of this MediaBulkUpdate. 
        :rtype: dict(str, object)
        """
        return self._attributes

    @attributes.setter
    def attributes(self, attributes):
        """
        Attribute values to bulk update an entity list.

        :param attributes: The attributes of this MediaBulkUpdate.
        :type: dict(str, object)
        """

        self._attributes = attributes

    @property
    def float_array(self):
        """
        Searches on `float_array` attributes.

        :return: The float_array of this MediaBulkUpdate. 
        :rtype: list[FloatArrayQuery]
        """
        return self._float_array

    @float_array.setter
    def float_array(self, float_array):
        """
        Searches on `float_array` attributes.

        :param float_array: The float_array of this MediaBulkUpdate.
        :type: list[FloatArrayQuery]
        """

        self._float_array = float_array

    @property
    def ids(self):
        """
        Array of media IDs to retrieve.

        :return: The ids of this MediaBulkUpdate. 
        :rtype: list[int]
        """
        return self._ids

    @ids.setter
    def ids(self, ids):
        """
        Array of media IDs to retrieve.

        :param ids: The ids of this MediaBulkUpdate.
        :type: list[int]
        """

        self._ids = ids

    @property
    def localization_ids(self):
        """
        Array of child localization IDs for which media should be retrieved.

        :return: The localization_ids of this MediaBulkUpdate. 
        :rtype: list[int]
        """
        return self._localization_ids

    @localization_ids.setter
    def localization_ids(self, localization_ids):
        """
        Array of child localization IDs for which media should be retrieved.

        :param localization_ids: The localization_ids of this MediaBulkUpdate.
        :type: list[int]
        """

        self._localization_ids = localization_ids

    @property
    def null_attributes(self):
        """
        Null a value in the attributes body

        :return: The null_attributes of this MediaBulkUpdate. 
        :rtype: list[str]
        """
        return self._null_attributes

    @null_attributes.setter
    def null_attributes(self, null_attributes):
        """
        Null a value in the attributes body

        :param null_attributes: The null_attributes of this MediaBulkUpdate.
        :type: list[str]
        """

        self._null_attributes = null_attributes

    @property
    def object_search(self):
        """

        :return: The object_search of this MediaBulkUpdate. 
        :rtype: AttributeOperationSpec
        """
        return self._object_search

    @object_search.setter
    def object_search(self, object_search):
        """

        :param object_search: The object_search of this MediaBulkUpdate.
        :type: AttributeOperationSpec
        """

        self._object_search = object_search

    @property
    def reset_attributes(self):
        """
        Reset an attribute to the default value specified in the Type object

        :return: The reset_attributes of this MediaBulkUpdate. 
        :rtype: list[str]
        """
        return self._reset_attributes

    @reset_attributes.setter
    def reset_attributes(self, reset_attributes):
        """
        Reset an attribute to the default value specified in the Type object

        :param reset_attributes: The reset_attributes of this MediaBulkUpdate.
        :type: list[str]
        """

        self._reset_attributes = reset_attributes

    @property
    def state_ids(self):
        """
        Array of child state IDs for which media should be retrieved.

        :return: The state_ids of this MediaBulkUpdate. 
        :rtype: list[int]
        """
        return self._state_ids

    @state_ids.setter
    def state_ids(self, state_ids):
        """
        Array of child state IDs for which media should be retrieved.

        :param state_ids: The state_ids of this MediaBulkUpdate.
        :type: list[int]
        """

        self._state_ids = state_ids

    @property
    def user_elemental_id(self):
        """
        Unique ID of the original user who created this. If permissions allow, will change the creating user to the one referenced by this elemental_id

        :return: The user_elemental_id of this MediaBulkUpdate. 
        :rtype: str
        """
        return self._user_elemental_id

    @user_elemental_id.setter
    def user_elemental_id(self, user_elemental_id):
        """
        Unique ID of the original user who created this. If permissions allow, will change the creating user to the one referenced by this elemental_id

        :param user_elemental_id: The user_elemental_id of this MediaBulkUpdate.
        :type: str
        """

        self._user_elemental_id = user_elemental_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, MediaBulkUpdate):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, MediaBulkUpdate):
            return True

        return self.to_dict() != other.to_dict()
